r"""

Collection of functions for performing numerical simulations on images
######################################################################

This module contains routines for performing simulations directly on images.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    simulations.drainage
    simulations.tortuosity_fd

"""

from ._drainage import *
from ._dns import *
from ._ibip import *
from ._ibip_gpu import *



from deprecated import deprecated
from porespy.simulations import tortuosity_fd as _tortuosity


__all__ = [
    'tortuosity',
]


@deprecated("The dns module has been renamed simulations and tortuosity was moved there")
def tortuosity(*args, **kwargs):
    return _tortuosity(*args, **kwargs)



import logging
import numpy as np
import scipy.ndimage as spim
from porespy.tools import extract_subsection, bbox_to_slices
from skimage.measure import mesh_surface_area
try:
    from skimage.measure import marching_cubes
except ImportError:
    from skimage.measure import marching_cubes_lewiner as marching_cubes
from skimage.morphology import skeletonize_3d, ball
from skimage.measure import regionprops
from skimage.measure._regionprops import RegionProperties
from pandas import DataFrame
from edt import edt


__all__ = [
    "regionprops_3D",
    "props_to_DataFrame",
    "prop_to_image",
]


logger = logging.getLogger(__name__)


def props_to_DataFrame(regionprops):
    r"""
    Create a ``pandas`` DataFrame containing all the scalar metrics for each
    region, such as volume, sphericity, and so on, calculated by
    ``regionprops_3D``.

    Parameters
    ----------
    regionprops : list
        This is a list of properties for each region that is computed
        by ``regionprops_3D``.  Because ``regionprops_3D`` returns data in
        the same ``list`` format as the ``regionprops`` function in **Skimage**
        you can pass in either.

    Returns
    -------
    DataFrame : Pandas DataFrame
        A Pandas DataFrame with each region corresponding to a row and each
        column corresponding to a key metric.  All the values for a given
        property (e.g. 'sphericity') can be obtained as
        ``val = df['sphericity']``.  Conversely, all the key metrics for a
        given region can be found with ``df.iloc[1]``.

    See Also
    --------
    prop_to_image
    regionprops_3d

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/props_to_DataFrame.html>`_
    to view online example.

    """
    # Parse the regionprops list and pull out all props with scalar values
    metrics = []
    reg = regionprops[0]
    for item in reg.__dir__():
        if not item.startswith('_'):
            try:
                if np.shape(getattr(reg, item)) == ():
                    metrics.append(item)
            except (TypeError, NotImplementedError, AttributeError):
                pass
    # Create a dictionary of all metrics that are simple scalar properties
    d = {}
    for i, k in enumerate(metrics):
        logger.info("Processing {k}")
        try:
            d[k] = np.array([r[k] for r in regionprops])
        except ValueError:  # pragma: no cover
            logger.error(f'Error encountered evaluating {k} so skipping it')
    # Create pandas data frame an return
    df = DataFrame(d)
    return df


def prop_to_image(regionprops, shape, prop):
    r"""
    Create an image with each region colored according the specified ``prop``,
    as obtained by ``regionprops_3d``.

    Parameters
    ----------
    regionprops : list
        This is a list of properties for each region that is computed
        by PoreSpy's ``regionprops_3D`` or Skimage's ``regionsprops``.
    shape : array_like
        The shape of the original image for which ``regionprops`` was obtained.
    prop : string
        The region property of interest.  Can be a scalar item such as 'volume'
        in which case the the regions will be colored by their respective
        volumes, or can be an image-type property such as 'border' or
        'convex_image', which will return an image composed of the sub-images.

    Returns
    -------
    image : ndarray
        An ndarray the same size as the original image, with each region
        represented by the values specified in ``prop``.

    See Also
    --------
    props_to_DataFrame
    regionprops_3d

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/prop_to_image.html>`_
    to view online example.

    """
    im = np.zeros(shape=shape)
    for r in regionprops:
        if prop == 'convex':
            mask = r.convex_image
        else:
            mask = r.image
        temp = mask * r[prop]
        s = bbox_to_slices(r.bbox)
        im[s] += temp
    return im


def regionprops_3D(im):
    r"""
    Calculates various metrics for each labeled region in a 3D image.

    This functions offers a few extras for 3D images that are not provided
    by the ``regionprops`` function in ``scikit-image``.

    Parameters
    ----------
    im : array_like
        An image containing at least one labeled region. If a boolean image
        is received than the ``True`` voxels are treated as a single region
        labeled ``1``.  Regions labeled 0 are ignored in all cases.

    Returns
    -------
    props : list
        An augmented version of the list returned by skimage's ``regionprops``.
        Information, such as ``volume``, can be found for region A using the
        following syntax: ``result[A-1].volume``.

        The returned list contains all the metrics normally returned by
        **skimage.measure.regionprops** plus the following:

        'slices'
            Slice indices into the image that can be used to extract the
            region

        'volume'
            Volume of the region in number of voxels.

        'bbox_volume'
            Volume of the bounding box that contains the region.

        'border'
            The edges of the region, found as the locations where the distance
            transform is 1.

        'inscribed_sphere'
            An image containing the largest sphere can can fit entirely inside
            the region.

        'surface_mesh_vertices'
            Obtained by applying the marching cubes algorithm on the region,
            AFTER first blurring the voxel image. This allows marching cubes
            more freedom to fit the surface contours.
            See also ``surface_mesh_simplices``

        'surface_mesh_simplices'
            This accompanies ``surface_mesh_vertices`` and together they can
            be used to define the region as a mesh.

        'surface_area'
            Calculated using the mesh obtained as described above, using the
            ``porespy.metrics.mesh_surface_area`` method.

        'sphericity'
            Defined as the ratio of the area of a sphere with the same volume
            as the region to the actual surface area of the region.

        'skeleton'
            The medial axis of the region obtained using the ``skeletonize_3D``
            method from **skimage**.

        'convex_volume'
            Same as convex_area, but translated to a more meaningful name.

    See Also
    --------
    snow_partitioning

    Notes
    -----
    Regions can be identified using a watershed algorithm, which can be a bit
    tricky to obtain desired results.  *PoreSpy* includes the SNOW algorithm,
    which may be helpful.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/regionprops_3D.html>`_
    to view online example.

    """
    results = regionprops(im)
    for i, obj in enumerate(results):
        a = results[i]
        b = RegionPropertiesPS(a.slice,
                               a.label,
                               a._label_image,
                               a._intensity_image,
                               a._cache_active)
        results[i] = b

    return results


class RegionPropertiesPS(RegionProperties):

    @property
    def mask(self):
        return self.image

    @property
    def slices(self):
        return self._slice

    @property
    def volume(self):
        return self.area

    @property
    def bbox_volume(self):
        mask = self.mask
        return np.prod(mask.shape)

    @property
    def border(self):
        return self.dt == 1

    @property
    def dt(self):
        mask = self.mask
        mask_padded = np.pad(mask, pad_width=1, mode='constant')
        temp = edt(mask_padded)
        return extract_subsection(temp, shape=mask.shape)

    @property
    def inscribed_sphere(self):
        dt = self.dt
        r = dt.max()
        inv_dt = edt(dt < r)
        return inv_dt < r

    @property
    def sphericity(self):
        vol = self.volume
        r = (3 / 4 / np.pi * vol)**(1 / 3)
        a_equiv = 4 * np.pi * r**2
        a_region = self.surface_area
        return a_equiv / a_region

    @property
    def skeleton(self):
        return skeletonize_3d(self.mask)

    @property
    def surface_area(self):
        mask = self.mask
        tmp = np.pad(np.atleast_3d(mask), pad_width=1, mode='constant')
        tmp = spim.convolve(tmp, weights=ball(1)) / 5
        verts, faces, norms, vals = marching_cubes(volume=tmp, level=0)
        self._surface_mesh_vertices = verts
        self._surface_mesh_simplices = faces
        area = mesh_surface_area(verts, faces)
        return area

    @property
    def surface_mesh_vertices(self):
        if not hasattr(self, '_surface_mesh_vertices'):
            _ = self.surface_area
        return self._surface_mesh_vertices

    @property
    def surface_mesh_simplices(self):
        if not hasattr(self, '_surface_mesh_simplices'):
            _ = self.surface_area
        return self._surface_mesh_simplices

    @property
    def convex_volume(self):
        return self.convex_area



import numpy as np
from porespy.tools import norm_to_uniform
import psutil


__all__ = [
    'fractal_noise',
]


def fractal_noise(
    shape,
    frequency=0.05,
    octaves=4,
    gain=0.5,
    mode='simplex',
    seed=None,
    cores=None,
    uniform=True,
):
    r"""
    Generate fractal noise which can be thresholded to create binary
    images with realistic structures across scales.

    Parameters
    ----------
    shape : array_like
        The size of the image to generate, can be 2D or 3D.
    frequency : scalar, default=0.05
        Controls the overall scale of the generated noise, with larger
        values giving smaller structures.
    octaves : int, default=4
        Controls the number of scales across which structures are
        generated, with larger values giving more scale.
    gain : scalar, default=0.5
        Controls the intensity of successively higher octaves. Values
        below 1.0 mean the higher octaves are less prominent.
    mode : string
        The type of noise to generate. Options are:

        ========== ==================================================================
        mode       description
        ========== ==================================================================
        'perlin'   Classic Perlin noise. For more information on ``perlin noise`` see
                   `here <https://en.wikipedia.org/wiki/Perlin_noise>`__.
        'simplex'  Updated Perlin noise for more realistic textures. For more
                   information on ``simplex noise`` see
                   `here <https://en.wikipedia.org/wiki/Simplex_noise>`__.
        'value'    Bilnear interpolation of white noise. For more information on
                   ``value noise`` see
                   `here <https://en.wikipedia.org/wiki/Value_noise>`__.
        'cubic'    Cubic interpolation of white noise. For more information on
                   ``cubic noise`` see
                   `here <https://github.com/jobtalle/CubicNoise>`__.
        ========== ==================================================================

    seed : int, optional
        The seed of the random number generator.  Using the same
        ``seed`` between runs will produce the same image.
    cores : int, optional
        The number of cores to use. This function uses ``pyfastnoisesimd``,
        which has implemented SIMD processing which can be spread across
        cores. The default is to use all cores.
    uniform : boolean, optional
        If ``True`` (default) the random values are converted to a
        uniform distribution between 0 and 1, otherwise the resulting image
        contains the unprocesssed values, which have a 'normal-esque'
        distribution centered on 0.

    Notes
    -----
    This function provides a simplified wrapper for the functions in
    the `pyfastnoisesimd <https://github.com/robbmcleod/pyfastnoisesimd>`_
    package. ``pyfastnoisesimd`` is itself a wrapper for a C-library
    called `FastNoiseSIMD <https://github.com/Auburn/FastNoiseSIMD>`_.
    To access the more elaborate functionality and options of these
    packages, explore the `pyfastnoisesimd documentation
    <https://pyfastnoisesimd.readthedocs.io/en/latest/overview.html>`_.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/fractal_noise.html>`_
    to view online example.

    """
    try:
        from pyfastnoisesimd import Noise, NoiseType, PerturbType
    except ModuleNotFoundError:
        raise ModuleNotFoundError("You need to install `pyfastnoisesimd` using"
                                  " `pip install pyfastnoisesimd`")
    if cores is None:
        cores = psutil.cpu_count(logical=False)
    if seed is None:
        seed = np.random.randint(2**31)
    perlin = Noise(numWorkers=cores)
    perlin.noiseType = getattr(NoiseType, f'{mode.capitalize()}Fractal')
    perlin.frequency = frequency
    perlin.fractal.octaves = octaves
    perlin.fractal.gain = gain
    perlin.fractal.lacunarity = 2
    perlin.perturb.perturbType = PerturbType.NoPerturb
    perlin.seed = seed
    result = perlin.genAsGrid(shape)
    if uniform:
        result = norm_to_uniform(result, scale=[0, 1])
    return result



r"""

Collection of functions for basic visualization of images
#########################################################

This module contains functions for quickly visualizing 3D images in 2D
views.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    visualization.bar
    visualization.imshow
    visualization.prep_for_imshow
    visualization.satn_to_movie
    visualization.satn_to_panels
    visualization.sem
    visualization.set_mpl_style
    visualization.show_3D
    visualization.show_mesh
    visualization.show_planes
    visualization.xray

"""

from ._views import *
from ._plots import *
from ._funcs import *



import numpy as np
import scipy.ndimage as spim
import math
from porespy import settings
from porespy.tools import extend_slice
from scipy.ndimage import zoom as zm
from skimage.morphology import ball
from porespy.tools import get_tqdm
tqdm = get_tqdm()


__all__ = [
    "diffusive_size_factor_AI",
    "diffusive_size_factor_DNS",
    "create_model",
    "find_conns",
]


def diffusive_size_factor_AI(regions, throat_conns, model,
                             g_train, voxel_size=1):
    '''
    Parameters
    ----------
    regions : ndarray
        A segmented 3D image of pore regions/a pair of two regions.
    throat_conns : array
        An Nt by 2 array containing the throat connections. The indices orders in
        throat_conns start from 0 to be consistent with network extraction method.
    model : tensorflow model
        The trained model to be used for prediction.
    g_train : array
        The training data distribution. This will be used for denormalizing
        the prediction.
    voxel_size : scalar, optional
        Voxel size of the image. The default is 1.

    Returns
    -------
    diff_size_factor : array
        An array of length conns containing diffusive size factor of the conduits
        in the segmented image (regions).

    Examples
    --------
    `Click here
    <https://porespy.org/examples/networks/reference/diffusive_size_factor_AI.html>`_
    to view online example.

    """
    '''
    import tensorflow as tf
    if g_train is None:
        raise ValueError("Training ground truth data must be given" +
                         "to be used for normalizing the test data")
    its = -1
    pairs = np.empty((len(throat_conns), 64, 64, 64), dtype='int32')
    diff_size_factor = []
    zm_ratios = []
    desc = 'Preparing images tensor'
    for i in tqdm(np.arange(len(throat_conns)), desc=desc, **settings.tqdm):
        cn = throat_conns[i]
        # crop two pore regions and label them as 1,2
        bb = _calc_bound_box_bi(regions, cn[0], cn[1])
        roi_crop = np.copy(regions[bb[0]:bb[3], bb[1]:bb[4], bb[2]:bb[5]])
        # label two pore regions as 1,2
        roi_masked = _create_labeled_pair(cn, roi_crop)
        # resize to AI input size
        resize_result = _resize_to_AI_input(roi_masked)
        roi_resized = resize_result['resized_im']
        zm_ratio = resize_result['zm_ratio']
        zm_ratios.append(zm_ratio)
        its = its+1
        pairs[its, :, :, :] = roi_resized.copy()
    test_data = _convert_to_tf_image_datatype(pairs, n_image=len(throat_conns))
    if len(throat_conns) < 16:
        batch_size = 1
    else:
        batch_size = 16
    test_steps = math.ceil(len(throat_conns) / batch_size)
    predictions = model.predict(test_data, steps=test_steps)
    predictions = np.squeeze(predictions)
    denorm_size_factor = _denorm_predict(predictions, g_train)
    g = denorm_size_factor * voxel_size * (1/np.array(zm_ratios))
    diff_size_factor = g
    if len(throat_conns) > 1:
        tf.keras.backend.clear_session()
    return diff_size_factor


def diffusive_size_factor_DNS(regions, throat_conns, voxel_size=1):
    """
    Calculates the diffusive size factor of pore to pore regions in
    a segmented image of porous material using finite difference method.
    Parameters
    ----------
    regions : ndarray
        A segmented 3D image of pore regions/a pair of two regions.
    throat_conns : array
        An Nt by 2 array containing the throat connections. The indices orders in
        throat_conns start from 0 to be consistent with network extraction method.
    voxel_size : scalar, optional
        Voxel size of the image. The default is 1.

    Returns
    -------
    diff_size_factor : array
        An array of length conns containing diffusive size factor of the conduits
        in the segmented image (regions).

    """
    DNS_size_factor = []
    desc = 'Preparing images and DNS calculations'
    settings.tqdm['disable'] = False
    for i in tqdm(np.arange(len(throat_conns)), desc=desc, **settings.tqdm):
        cn = throat_conns[i]
        # crop two pore regions and label them as 1,2
        bb = _calc_bound_box_bi(regions, cn[0], cn[1])
        roi_crop = np.copy(regions[bb[0]:bb[3], bb[1]:bb[4], bb[2]:bb[5]])
        # label two pore regions as 1,2
        roi_masked = _create_labeled_pair(cn, roi_crop)
        DNS_size_factor.append(_calc_g_val(roi_masked))
    diff_size_factor = np.array(DNS_size_factor) * voxel_size
    return diff_size_factor


def _calc_g_val(im):
    '''
    Calculates the diffusive size factor of conduit image (ROI)
    using finite difference method. The finite difference nodes
    are created using OpenPNM's CubicTemplate method.
    Parameters
    ----------
    im : ndarray
        3D image of a pair of pore to pore regions (conduit)
    Returns
    -------
    g : scalar
        Diffusive size factor of the conduit.
    '''
    import openpnm as op
    c1 = 20
    c2 = 10
    im = np.copy(im)
    results = _find_conns_roi_info(im)
    centroids = results['p_coords']
    p_dia_local = results['p_dia_local']
    # create a mask where solid phase==True to trim the nodes
    # in the finite difference nodes that are located in solid/not ROI
    mask1 = np.array(np.where(im[:] == 1, im[:], 0),
                     dtype=bool)
    mask2 = np.array(np.where(im[:] == 2, im[:], 0),
                     dtype=bool)
    mask1 = np.reshape(mask1, mask1.size)
    mask2 = np.reshape(mask2, mask2.size)
    mask = ~(mask1+mask2)
    n = im.shape
    meds = op.network.Cubic(shape=[n[0], n[1], n[2]], spacing=1)
    meds['pore.region1'] = mask1.copy()
    meds['pore.region2'] = mask2.copy()
    # trim nodes that are located in solid phase/not ROI
    op.topotools.trim(meds, pores=mask)
    meds.add_model(propname='pore.cluster_number',
                   model=op.models.network.cluster_number)
    meds.add_model(propname='pore.cluster_size',
                   model=op.models.network.cluster_size)
    cluster_size = np.max(meds['pore.cluster_size'])
    trim_pores = meds['pore.cluster_size'] < cluster_size
    op.topotools.trim(network=meds, pores=trim_pores)
    phss = op.phase.Phase(network=meds)
    # A diffusive conductance of 1 ensures a finite difference approach
    # where each node is located at the corner of each voxel
    phss['throat.diffusive_conductance'] = 1
    algs = op.algorithms.FickianDiffusion(network=meds,
                                          phase=phss)
    # find centroid of each pore region in the finite difference nodes
    pr1 = closest_node(centroids[0], meds['pore.coords'])
    pr2 = closest_node(centroids[1], meds['pore.coords'])
    algs.set_value_BC(pores=pr1, values=c1)
    algs.set_value_BC(pores=pr2, values=c2)
    algs.run()
    # calculate average concentrations within inscribed spheres
    r1 = p_dia_local[0]/2
    r2 = p_dia_local[1]/2
    if np.round(r1) == 0:
        # This prevent error in find_nearby_pores for narrow regions
        # If the region is narrow, use the entire region for average concentration
        c1_avr = algs['pore.concentration'][meds['pore.region1']].mean()
    else:
        # use the inscribed sphere within the region for average concentration
        pores1 = meds.find_nearby_pores(pr1, r=np.round(r1), flatten=True,
                                        include_input=True)
        pores1 = np.append(pores1, pr1)
        pores1 = np.unique(pores1)
        c1_avr = algs['pore.concentration'][pores1].mean()
    if np.round(r2) == 0:
        c2_avr = algs['pore.concentration'][meds['pore.region2']].mean()
    else:
        pores2 = meds.find_nearby_pores(pr2, r=np.round(r2), flatten=True,
                                        include_input=True)
        pores2 = np.append(pores2, pr2)
        pores2 = np.unique(pores2)
        c2_avr = algs['pore.concentration'][pores2].mean()
    g = abs(algs.rate(pores=pr1)[0]/(c1_avr-c2_avr))
    return g


def closest_node(extracted_nodes, fd_nodes):
    """
    Finds the indice of a node in the finite difference
    nodes that locates closest to the centroid point of a pore region.

    Parameters
    ----------
    extracted_nodes : array
        An array of the coordinate of the centroid point of a pores region.
    fd_nodes : array
        An array of the coordinate of all nodes in the finite difference
        nodes.

    Returns
    -------
    scalar
        The indice of the nearest finite difference node to the
        centroid of a pore region.

    """
    fd_nodes = np.asarray(fd_nodes)
    dist = np.sum((fd_nodes - extracted_nodes)**2, axis=1)
    return int(np.argmin(dist))


def _find_conns_roi_info(im):
    '''
    Finds the connections list, coordinates of pores centroids and
    their inscribed sphere's diameter. These values are necessary to be known
    for applying the finite difference method.
    Parameters
    ----------
    im : ndarray
        A segmented image of a porous medium.
    Returns
    -------
    A dictionary of info:
    t_conns : array
        An Nt by 2 array containing the throats' connections in the segmented image.
    p_coords : array
        An Np by 3 array  containing the pores centroids coordinates in the
        segmented image.
    p_dia_local : array
        An Np size array  containing the pores inscribed diameter in the
        segmented image.
    '''
    struc_elem = ball
    slices = spim.find_objects(im)
    Ps = np.arange(1, np.amax(im)+1)
    p_dia_local = np.zeros((len(Ps), ), dtype=float)
    p_coords = np.zeros((len(Ps), im.ndim), dtype=float)
    t_conns = []
    desc = 'Getting ROI info'
    settings.tqdm['disable'] = True
    for i in tqdm(Ps, desc=desc, **settings.tqdm):
        pore = i - 1
        if slices[pore] is None:
            continue
        s = extend_slice(slices[pore], im.shape)
        sub_im = im[s]
        pore_im = sub_im == i
        # additional info to find centroids and inscribed_diam
        padded_mask = np.pad(pore_im, pad_width=1, mode='constant')
        pore_dt = spim.distance_transform_edt(padded_mask)
        s_offset = np.array([i.start for i in s])
        p_coords[pore, :] = spim.center_of_mass(pore_im) + s_offset
        p_dia_local[pore] = (2*np.amax(pore_dt)) - np.sqrt(3)
        im_w_throats = spim.binary_dilation(input=pore_im, structure=struc_elem(1))
        im_w_throats = im_w_throats*sub_im
        Pn = np.unique(im_w_throats)[1:] - 1
        for j in Pn:
            if j > pore:
                t_conns.append([pore, j])
    results = {
        't_conns': t_conns,
        'p_coords': p_coords,
        'p_dia_local': p_dia_local
    }
    return results


def _calc_bound_box_bi(regions, pore_1, pore_2):
    '''
    Parameters
    ----------
    regions : ndarray
        3D image of the segmented regions from which the bounding box of
        a local pore to pore region will be extracted.
    pore_1 : scalar
        Label of pore region1 in a pair of pore to pore connection (throat)
    pore_2 : scalar
        Label of pore region1 in a pair of pore to pore connection (throat)

    Returns
    -------
    b_box : array
        An array containing the coordinates of the bounding box that covers
        region pore_1 and region pore_2. The orders of coordinates are:
        x_min, y_min, z_min, x_max, y_max, z_max.

    '''
    slice_x, slice_y, slice_z = spim.find_objects(regions == pore_1+1)[0]
    slice_x2, slice_y2, slice_z2 = spim.find_objects(regions == pore_2+1)[0]
    min_box = [min(slice_x.start, slice_x2.start),
               min(slice_y.start, slice_y2.start),
               min(slice_z.start, slice_z2.start)]
    max_box = [max(slice_x.stop, slice_x2.stop), max(slice_y.stop, slice_y2.stop),
               max(slice_z.stop, slice_z2.stop)]
    b_box = np.hstack([min_box, max_box])
    return b_box


def _resize_to_AI_input(im):
    '''
    Parameters
    ----------
    im : ndarray
        3D image of a pair of pore to pore regions (conduit) cropped from
        segmented image of a porous medium.

    Returns
    -------
    result : dict
        Contains Resized image of the cropped regions and its zoom ratio.
        If original image(im) is of shape (n,n,n) resizing includes one step
        of zoom to a (64,64,64) image. Otherwise, the original image
        will be first zero padded to its maximum dimension (nmax,nmax,nmax)
        before zoom step.

    '''
    if len(np.unique(im.shape)) != 1:
        x, y, z = np.shape(im)
        max_size = max([x, y, z])
        x_diff, y_diff, z_diff = max_size-np.array([x, y, z])
        XB = int(np.round(x_diff/2))  # padding before axis x
        XA = int(x_diff-XB)  # padding after axis x
        YB = int(np.round(y_diff/2))
        YA = int(y_diff-YB)
        ZB = int(np.round(z_diff/2))
        ZA = int(z_diff-ZB)
        im = np.pad(im, ((XB, XA), (YB, YA), (ZB, ZA)), 'constant',
                    constant_values=0)
    zm_ratio = 64/im.shape[0]
    resized_im = zm(im, zoom=[zm_ratio, zm_ratio, zm_ratio], order=0)
    result = {'zm_ratio': zm_ratio, 'resized_im': resized_im}
    return result


def find_conns(im):
    '''
    Parameters
    ----------
    im : ndarray
        A segmented image of a porous medium.

    Returns
    -------
    t_conns : array
        An Nt by 2 addat containing the throats' connections in the
          segmented image.
    '''
    struc_elem = ball
    slices = spim.find_objects(im)
    Ps = np.arange(1, np.amax(im)+1)
    t_conns = []
    d = 'Finding neighbouring regions'
    settings.tqdm['disable'] = True
    for i in tqdm(Ps, desc=d, **settings.tqdm):
        pore = i - 1
        if slices[pore] is None:
            continue
        s = extend_slice(slices[pore], im.shape)
        sub_im = im[s]
        pore_im = sub_im == i
        im_w_throats = spim.binary_dilation(input=pore_im, structure=struc_elem(1))
        im_w_throats = im_w_throats*sub_im
        Pn = np.unique(im_w_throats)[1:] - 1
        for j in Pn:
            if j > pore:
                t_conns.append([pore, j])
    return t_conns


def _create_labeled_pair(cn, im):
    '''
    Parameters
    ----------
    cn : array
        An array containing two elements [region1,region2] that are labels of
        pore regions in a pair (im).
    im : ndarray
        3D image of a pair of neighbouring pores.

    Returns
    -------
    roi_masked : ndarray
        labeled image of two pore regions (im) where solid, region1, and region2 are
        labeled as 0, 1, and 2, respectively.

    '''
    # create labeled image for AI : labels (0,1,2) for (solid,pore1,pore2)
    roi_crop = np.copy(im)
    mask1 = np.array(np.where(roi_crop == cn[0]+1, roi_crop, 0),
                     dtype=bool)
    mask2 = np.array(np.where(roi_crop == cn[1]+1, roi_crop, 0),
                     dtype=bool)
    roi_masked = 1*mask1 + 2*mask2
    return roi_masked


def _convert_to_tf_image_datatype(pair, n_image):
    '''

    Parameters
    ----------
    pair : ndarray
        Labeled pore to pore regions.
    n_image : scalar
        Number of pairs of conduit regions.

    Returns
    -------
    test_data : tensorflow data
        Tensorflow data type test data. The test data may include 1 or more pairs
        of conduit images.

    '''
    import tensorflow as tf
    # create a tensor of size (n_image, pair_shape)
    data_ims = np.zeros(shape=(n_image, 64, 64, 64, 1))
    data = np.expand_dims(pair, axis=-1)
    if n_image == 1:
        data_ims[0, :, :, :] = data
        BS = 1
    else:
        data_ims = data
        if n_image < 16:
            BS = 1
        else:
            BS = 16
    # create test_data as a tensorflow dataset type
    AUTOTUNE = tf.data.experimental.AUTOTUNE
    test_data = tf.data.Dataset.from_tensor_slices(data_ims)
    test_data = test_data.cache()
    test_data = test_data.batch(BS)   # batch size = 1
    test_data = test_data.prefetch(AUTOTUNE)
    return test_data


def _denorm_predict(prediction, g_train):
    '''
    Parameters
    ----------
    prediction : array
        Normalized predicted diffusive size factors for conduit images.
    g_train : array
        Diffusive size factors used in training AI.

    Returns
    -------
    denorm : array
        Denormalized predicted diffusive size factor for conduit images.

    '''
    from sklearn import preprocessing
    scaler = preprocessing.MinMaxScaler(feature_range=(0, 1))
    train_N = scaler.fit_transform(g_train.reshape(-1, 1))
    denorm = scaler.inverse_transform(X=prediction.reshape(-1, 1))
    denorm = np.squeeze(denorm)
    return denorm


def _id_block(x, filters, kernel_size):
    from tensorflow.keras import layers as ly
    from tensorflow.keras.initializers import glorot_uniform
    f1, f2, f3 = filters
    k = kernel_size
    x_orig = x

    x = ly.Conv3D(filters=f1, kernel_size=(1, 1, 1),
                  strides=(1, 1, 1), padding='valid',
                  kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.BatchNormalization(axis=-1)(x)
    x = ly.Activation('relu')(x)

    x = ly.Conv3D(filters=f2, kernel_size=(k, k, k),
                  strides=(1, 1, 1), padding='same',
                  kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.BatchNormalization(axis=-1)(x)
    x = ly.Activation('relu')(x)

    x = ly.Conv3D(filters=f3, kernel_size=(1, 1, 1),
                  strides=(1, 1, 1), padding='valid',
                  kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.BatchNormalization(axis=-1)(x)

    x = ly.Add()([x, x_orig])
    x = ly.Activation('relu')(x)

    return x


def _conv_block(x, filters, kernel_size, stride):
    from tensorflow.keras import layers as ly
    from tensorflow.keras.initializers import glorot_uniform
    f1, f2, f3 = filters
    k = kernel_size
    s = stride
    x_orig = x

    x = ly.Conv3D(filters=f1, kernel_size=(1, 1, 1),
                  strides=(s, s, s), padding='valid',
                  kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.BatchNormalization(axis=-1)(x)
    x = ly.Activation('relu')(x)

    x = ly.Conv3D(filters=f2, kernel_size=(k, k, k),
                  strides=(1, 1, 1), padding='same',
                  kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.BatchNormalization(axis=-1)(x)
    x = ly.Activation('relu')(x)

    x = ly.Conv3D(filters=f3, kernel_size=(1, 1, 1),
                  strides=(1, 1, 1), padding='valid',
                  kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.BatchNormalization(axis=-1)(x)

    x_shortcut = ly.Conv3D(filters=f3, kernel_size=(1, 1, 1),
                           strides=(s, s, s), padding='valid',
                           kernel_initializer=glorot_uniform(seed=0))(x_orig)
    x_shortcut = ly.BatchNormalization(axis=-1)(x_shortcut)

    x = ly.Add()([x, x_shortcut])
    x = ly.Activation('relu')(x)

    return x


def _resnet3d(input_shape=(64, 64, 64, 1)):
    from tensorflow.keras import layers as ly
    from tensorflow.keras.initializers import glorot_uniform
    from tensorflow.keras.models import Model
    x_in = ly.Input(shape=input_shape)

    x = ly.ZeroPadding3D(padding=(3, 3, 3))(x_in)

    # stage 1
    x = ly.Conv3D(filters=64, kernel_size=(7, 7, 7), strides=(2, 2, 2),
                  kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.BatchNormalization(axis=-1)(x)
    x = ly.Activation('relu')(x)
    x = ly.MaxPooling3D(pool_size=(3, 3, 3), strides=(2, 2, 2))(x)

    # stage 2
    x = _conv_block(x, filters=(64, 64, 256), kernel_size=3, stride=1)
    x = _id_block(x, filters=(64, 64, 256), kernel_size=3)
    x = _id_block(x, filters=(64, 64, 256), kernel_size=3)

    # stage 3
    x = _conv_block(x, filters=(128, 128, 512), kernel_size=3, stride=2)
    x = _id_block(x, filters=(128, 128, 512), kernel_size=3)
    x = _id_block(x, filters=(128, 128, 512), kernel_size=3)
    x = _id_block(x, filters=(128, 128, 512), kernel_size=3)

    # stage 4
    x = _conv_block(x, filters=(256, 256, 1024), kernel_size=3, stride=2)
    x = _id_block(x, filters=(256, 256, 1024), kernel_size=3)
    x = _id_block(x, filters=(256, 256, 1024), kernel_size=3)
    x = _id_block(x, filters=(256, 256, 1024), kernel_size=3)
    x = _id_block(x, filters=(256, 256, 1024), kernel_size=3)
    x = _id_block(x, filters=(256, 256, 1024), kernel_size=3)

    # stage 5
    x = _conv_block(x, filters=(512, 512, 2048), kernel_size=3, stride=2)
    x = _id_block(x, filters=(512, 512, 2048), kernel_size=3)
    x = _id_block(x, filters=(512, 512, 2048), kernel_size=3)

    # average pooling
    x = ly.AveragePooling3D(pool_size=(2, 2, 2))(x)

    # output layer
    x = ly.Flatten()(x)

    x = ly.Dense(units=512, kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.Activation(activation='relu')(x)
    x = ly.Dropout(rate=0.5)(x)

    x = ly.Dense(units=1, kernel_initializer=glorot_uniform(seed=0))(x)
    x = ly.Activation(activation='linear')(x)

    # create model
    model = Model(inputs=x_in, outputs=x, name='resnet3d')

    return model


def create_model():
    '''
    Returns
    -------
    model : tensorflow model
        ResNet50 model built using convolutional and identity blocks.

    '''
    from tensorflow.keras.optimizers import Adam
    model = _resnet3d()
    model.compile(loss='mse', optimizer=Adam(lr=1e-4), metrics=['mse'])
    return model



r"""

imagej
######

**Access ImageJ Filters**

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    filters.imagej.imagej_wrapper
    filters.imagej.imagej_plugin

"""

from ._funcs import imagej_wrapper
from ._funcs import imagej_plugin



import logging
import numpy as np
from porespy.tools import _insert_disks_at_points, find_bbox


__all__ = [
    'spheres_from_coords',
]


logger = logging.getLogger(__name__)


def spheres_from_coords(df, mode='contained', smooth=False):
    r"""
    Generates a sphere packing given a list of centers and radii

    Parameters
    ----------
    df : DataFrame or dict
        The X, Y, Z center coordinates, and radius R of each sphere in the packing
        should be stored in their own column or key. The units should be in voxels.
        See ``Notes`` for more detail on how this should be formatted. If one of
        the dimensions has all 0's then a 2D image is generated.
    smooth : bool
        Indicates if spheres should be smooth or have the single pixel bump on
        each face.
    mode : str
        How edges are handled. Options are:

        ============= ==============================================================
        mode          description
        ============= ==============================================================
        'contained'   (default) All spheres are fully contained within the image,
                      meaning the image is padded beyond extents of the given
                      coordinates by the maximum radius of the given sphere radii
                      to ensure they all fit.
        'extended'    Spheres extend beyond the edge of the image. In this mode
                      the image is only as large enough to hold the given coordinates
                      so the spheres may extend beyond the image boundary.
        ============= ==============================================================

    Returns
    -------
    spheres : ndarray
        A numpy ndarray of ``True`` values indicating the spheres and
        ``False`` elsewhere. The size of the returned image will be large enough
        to fit all the spheres plus the radius of the largest sphere.

    Notes
    -----
    The input data should be in column format as a dictionary of 1D
    *numpy* arrays like this:

    .. code::

        d['X'] = np.array([1, 1, 2])
        d['Y'] = np.array([1, 3, 2])
        d['Z'] = np.array([1, 1, 1])
        d['R'] = np.array([0.5, 0.7, 0.6])

    Or a *pandas* ``DataFrame` like this:

    ==== ==== ==== ==== ====
    ID   'X'  'Y'  'Z'  'R'
    ==== ==== ==== ==== ====
    0    1    1    1    0.5
    1    1    3    1    0.7
    2    2    2    1    0.6
    ==== ==== ==== ==== ====

    Or a numpy N-by-3(or 4) array like this:

    .. code::

        array([[1, 1, 1, 0.5],
               [1, 3, 1, 0.7],
               [2, 2, 1, 0.6]])

    Note that in all cases if the radius is not given that it is assumed to be
    a single pixel

    """
    # Convert dict to have standard column names
    cols = ['X', 'Y', 'Z', 'R']
    if hasattr(df, 'keys'):
        for i, k in enumerate(df.keys()):
            if k.upper()[0] in cols:
                df[k.upper()[0]] = np.array(df[k])
    else:  # Assume it's a numpy array
        import pandas as pd
        df = pd.DataFrame(df, columns=cols[:df.shape[1]])

    if 'R' not in df.keys():
        df['R'] = np.ones_like(df['X'])

    # Correct for any negative coordinates
    for ax in ['X', 'Y', 'Z']:
        if np.any(df[ax] < 0):
            df[ax] -= df[ax].min()

    r = np.array(np.around(df['R'], decimals=0)).astype(int)
    x = np.array(np.around(df['X'], decimals=0)).astype(int)
    y = np.array(np.around(df['Y'], decimals=0)).astype(int)
    z = np.array(np.around(df['Z'], decimals=0)).astype(int)

    if mode == 'contained':
        x += r.max()
        y += r.max()
        z += r.max()
        shape = np.ceil([x.max() + 1 + r.max(),
                         y.max() + 1 + r.max(),
                         z.max() + 1 + r.max()]).astype(int)
        crds = np.vstack([x, y, z]).T
    elif mode == 'extended':
        shape = np.ceil([x.max() + 1,
                         y.max() + 1,
                         z.max() + 1]).astype(int)
        crds = np.vstack([x, y, z]).T

    mask = np.all(crds == crds[0, :], axis=0)
    if np.any(mask):
        crds[:, mask] = 0
        shape[mask] = 1

    logger.info(f"Inserting spheres into image of size {shape}")
    im = np.zeros(shape, dtype=bool)
    im = _insert_disks_at_points(
        im,
        coords=crds.T,
        radii=r,
        v=True,
        smooth=smooth,
    )
    logger.info("Sphere insertion complete, performing postprocessing")
    im = im.squeeze()
    bbox = find_bbox(im)
    im = im[bbox]
    return im



import numpy as np
from porespy.tools import make_contiguous
from scipy.stats import rankdata


__all__ = [
    'size_to_seq',
    'size_to_satn',
    'seq_to_satn',
    'pc_to_satn',
    'satn_to_seq',
]


def size_to_seq(size, im=None, bins=None, mode='drainage'):
    r"""
    Converts an image of invasion size values into invasion sequence values

    Parameters
    ----------
    size : ndarray
        The image containing invasion size in each voxel. Values of 0 are
        assumed to be solid (if ``im`` is not given) and values of -1 are
        assumed to be uninvaded.
    im : ndarray, optional
        A binary image of the porous media, with ``True`` indicating the
        void space and ``False`` indicating the solid phase. If not given
        then it is assumed that the solid is identified as ``size == 0``.
    bins : array_like or int (optional)
        The bins to use when converting sizes to sequence.  The default is
        to create 1 bin for each unique value in ``size`` (except for -1 and 0).
        If an **int** is supplied, it is interpreted as the number of bins between 1
        and the maximum value in ``size``.  If an array is supplied it is used as
        the bins directly.
    mode : str
        Controls how the sizes are converted to a sequence. The options are:

        ============= ==============================================================
        `mode`        Description
        ============= ==============================================================
        'drainage'    The sizes are assumed to have been filled from largest to
                      smallest, ignoring 0's and -1's
        'imbibition'  The sizes are assumed to have been filled from smallest to
                      largest, ignoring 0's and -1's
        ============= ==============================================================

    Returns
    -------
    seq : ndarray
        An ndarray the same shape as ``size`` with invasion size values
        replaced by the invasion sequence, according to the specified `mode`.
        Any uninvaded voxels, indicated by -1 in ``size`` will be indicated by
        -1 in ``seq``.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/size_to_seq.html>`_
    to view online example.

    """
    if im is None:
        solid = size == 0
    else:
        solid = im == 0
    uninvaded = size == -1
    if bins is None:
        bins = np.unique(size)
    elif isinstance(bins, int):
        bins = np.linspace(0, size.max(), bins)
    vals = np.digitize(size, bins=bins, right=True)
    if mode.startswith('im'):
        vals[solid] = 0
        vals[uninvaded] = -1
        vals = make_contiguous(vals, mode='symmetric')
    if mode.startswith('dr'):
        vals = make_contiguous(vals, mode='symmetric')
        vals = vals.max() + 1 - vals
        vals[solid] = 0
        vals[uninvaded] = -1
    return vals


def size_to_satn(size, im=None, bins=None, mode='drainage'):
    r"""
    Converts an image of invasion size values into non-wetting phase saturations.

    Parameters
    ----------
    size : ndarray
        The image containing invasion size values in each voxel. Solid
        should be indicated as 0's and uninvaded voxels as -1.
    im : ndarray, optional
        A binary image of the porous media, with ``True`` indicating the
        void space and ``False`` indicating the solid phase. If not given
        then it is assumed that the solid is identified as ``size == 0``.
    bins : array_like or int (optional)
        The bins to use when converting sizes to saturation.  The default is
        to create 1 bin for each unique value in ``size``.  If an **int**
        is supplied it is interpreted as the number of bins between 0 and the
        maximum value in ``size``.  If an array is supplied it is used as
        the bins directly.
    mode : str
        Controls how the sizes are converted to saturations. The options are:

        ============= ==============================================================
        `mode`        Description
        ============= ==============================================================
        'drainage'    The sizes are assumed to have been filled from largest to
                      smallest, ignoring 0's and -1's
        'imbibition'  The sizes are assumed to have been filled from smallest to
                      largest, ignoring 0's and -1's
        ============= ==============================================================

    Returns
    -------
    satn : ndarray
        An ndarray the same shape as ``size`` but with size values replaced
        by the fraction of void space invaded at each size, according to the
        specified `mode`. Solid voxels and uninvaded voxels are represented by 0
        and -1, respectively.

    Notes
    -----
    If any ``-1`` values are present in `size` the maximum saturation will be less
    than 1.0 since this means that not all wetting phase was displaced.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/size_to_satn.html>`_
    to view online example.
    """
    if bins is None:
        bins = np.unique(size[size > 0])
    elif isinstance(bins, int):
        bins = np.linspace(0, size.max(), bins)
    if im is None:
        im = ~(size == 0)
    void_vol = im.sum()
    satn = -np.ones_like(size, dtype=float)
    if mode.startswith('im'):
        for r in bins:
            hits = (size <= r) * (size > 0)
            temp = hits.sum()/void_vol
            satn[hits * (satn == -1)] = temp
    elif mode.startswith('dr'):
        for r in bins[-1::-1]:
            hits = (size >= r) * (size > 0)
            temp = hits.sum()/void_vol
            satn[hits * (satn == -1)] = temp
    satn *= (im > 0)
    return satn


def seq_to_satn(seq, im=None, mode='drainage'):
    r"""
    Converts an image of invasion sequence values to non-wetting phase saturation
    values.

    Parameters
    ----------
    seq : ndarray
        The image containing invasion sequence values in each voxel. Solid
        should be indicated as 0's and uninvaded voxels as -1.
    im : ndarray, optional
        A binary image of the porous media, with ``True`` indicating the
        void space and ``False`` indicating the solid phase. If not given
        then it is assumed that the solid is identified as ``seq == 0``.
    mode : str
        Controls how the sequences are converted to saturations. The options are:

        ============= ==============================================================
        `mode`        Description
        ============= ==============================================================
        'drainage'    The saturation is assumed to increase with increasing sequence
        'imbibition'  The saturation is assumed to decrease with increasing sequence
        ============= ==============================================================

    Returns
    -------
    satn : ndarray
        An ndarray the same shape as ``seq`` but with sequence values replaced
        by the fraction of void space invaded at the sequence number, accounting
        for the specified `mode`. Solid voxels and uninvaded voxels are represented
        by 0 and -1, respectively.

    Notes
    -----
    If any ``-1`` values are present in `seq` the maximum saturation will be less
    than 1.0 since this means that not all wetting phase was displaced.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/seq_to_satn.html>`_
    to view online example.
    """
    seq = np.copy(seq).astype(int)
    if im is None:
        solid_mask = seq == 0
    else:
        solid_mask = im == 0
    uninvaded_mask = seq == -1  # Store uninvaded locations
    seq[seq <= 0] = 0  # Set uninvaded to solid for next steps
    if mode.startswith('im'):
        seq = seq.max() - seq + 1
        seq[solid_mask] = 0
        seq[uninvaded_mask] = 0
    seq = rankdata(seq, method='dense') - 1
    b = np.bincount(seq)
    if (solid_mask.sum(dtype=np.int64) > 0) or \
            (uninvaded_mask.sum(dtype=np.int64) > 0):
        b[0] = 0
    c = np.cumsum(b)
    seq = np.reshape(seq, solid_mask.shape)
    satn = c[seq]/(seq.size - solid_mask.sum(dtype=np.int64))
    satn[solid_mask] = 0
    satn[uninvaded_mask] = -1
    return satn


def pc_to_satn(pc, im, mode='drainage'):
    r"""
    Converts an image of capillary entry pressures to saturation values

    Parameters
    ----------
    pc : ndarray
        A Numpy array with the value in each voxel indicating the capillary
        pressure at which it was invaded. In order to accommodate the
        possibility of both positive and negative capillary pressure values,
        uninvaded voxels should be indicated by ``+inf`` and residual phase
        by ``-inf``. Solid vs void phase is defined by ``im`` which is
        mandatory.
    im : ndarray
        A Numpy array with ``True`` values indicating the void space
    mode : str
        Controls how the pressures are converted to sequence. The options are:

        ============= ==============================================================
        `mode`        Description
        ============= ==============================================================
        'drainage'    The pressures are assumed to have been filled from smallest to
                      largest, ignoring +/- infs
        'imbibition'  The pressures are assumed to have been filled from largest to
                      smallest, ignoring +/- infs
        ============= ==============================================================

    Returns
    -------
    satn : ndarray
        A Numpy array the same shape as `pc`, with each voxel value indicating
        the global saturation at which it was invaded, according to the specified
        `mode`.

    Notes
    -----
    If any ``+inf`` values are present the maximum saturation will be less than
    1.0 since not all wetting phase was displaced.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/pc_to_satn.html>`_
    to view online example.

    """
    a = np.digitize(pc, bins=np.unique(pc))
    a[~im] = 0
    a[np.where(pc == np.inf)] = -1
    satn = seq_to_satn(seq=a, im=im, mode=mode)
    return satn


def satn_to_seq(satn, im=None, mode='drainage'):
    r"""
    Converts an image of nonwetting phase saturations to invasion sequence
    values

    Parameters
    ----------
    satn : ndarray
        A Numpy array with the value in each voxel indicating the global
        saturation at the point it was invaded. -1 indicates a voxel that
        not invaded, and 0 indicates solid phase.
    im : ndarray
        A Numpy array with ``True`` values indicating the void space.
    mode : str
        Controls how the saturations are converted to sequence. The options are:

        ============= ==============================================================
        `mode`        Description
        ============= ==============================================================
        'drainage'    The pressures are assumed to have been filled from smallest to
                      largest, ignoring 0's and -1's
        'imbibition'  The sizes are assumed to have been filled from largest to
                      smallest, ignoring 0's and -1's
        ============= ==============================================================

    Returns
    -------
    seq : ndarray
        A Numpy array the same shape as `satn` with each voxel value indicating
        the sequence in which it was invaded, according to the specified `mode`.
        Solid voxels are indicated by 0 and uninvaded by -1.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/satn_to_seq.html>`_
    to view online example.

    """
    if im is None:
        im = satn > 0
    uninvaded = satn == -1
    values = np.unique(satn)
    seq = np.digitize(satn, bins=values)
    # Set uninvaded by to -1
    seq[satn == -1] = -1
    # Set solids back to 0
    seq[~im] = 0
    # Ensure values are contiguous while keeping -1 and 0
    seq = make_contiguous(im=seq, mode='symmetric')
    if mode.startswith('im'):
        seq = (seq.max() + 1) - seq
        seq[~im] = 0
    seq[uninvaded] = -1
    return seq



r"""

Collection of helper functions for manipulating images
######################################################

This module contains a variety of functions for manipulating images in
ways that do NOT return a modified version of the original image.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    tools.Results
    tools.align_image_with_openpnm
    tools.bbox_to_slices
    tools.extend_slice
    tools.extract_cylinder
    tools.extract_regions
    tools.extract_subsection
    tools.find_outer_region
    tools.get_border
    tools.get_planes
    tools.get_tqdm
    tools.in_hull
    tools.insert_cylinder
    tools.insert_sphere
    tools.isolate_object
    tools.make_contiguous
    tools.marching_map
    tools.mesh_region
    tools.norm_to_uniform
    tools.overlay
    tools.ps_ball
    tools.ps_disk
    tools.ps_rect
    tools.ps_round
    tools.randomize_colors
    tools.recombine
    tools.sanitize_filename
    tools.show_docstring
    tools.subdivide
    tools.unpad

"""

from ._funcs import *
from ._utils import *
from ._sphere_insertions import *


def _get_version():
    from porespy.__version__ import __version__ as ver
    suffix = ".dev0"
    if ver.endswith(suffix):
        ver = ver[:-len(suffix)]
    return ver



r"""

Collection of functions for obtaining network representations of images
#######################################################################

Contains functions for analysing images as pore networks.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    networks.add_boundary_regions
    networks.diffusive_size_factor_AI
    networks.generate_voxel_image
    networks.label_boundaries
    networks.label_phases
    networks.map_to_regions
    networks.maximal_ball_wrapper
    networks.regions_to_network
    networks.snow2

"""

from ._funcs import *
from ._maximal_ball import *
from ._getnet import *
from ._snow2 import *
from ._utils import *
from ._size_factors import *



import logging
import os
import time
import psutil
import subprocess
import numpy as np


__all__ = [
    "maximal_ball_wrapper",
]


logger = logging.getLogger(__name__)


def maximal_ball_wrapper(im, prefix, path_to_exe, voxel_size=1e-6):
    r"""
    Implementing a maximal ball algorithm on an image

    Parameters
    ----------
    im : ndarray
        The image of the porous material.
    prefix : string
        The prefix to append to the filenames (i.e. 'prefix_node1.dat')
    path_to_exe : string
        Path to the maximal ball .exe file (pnextract.exe). See Notes
    voxel_size : scalar
        The size of a voxel side in preferred units.  The default is 1e-6, so the
        user can apply the scaling to the returned results after the fact.

    Notes
    -----
    outputs four DAT files:
        prefix_link1, prefix_link2, prefix_node1, prefix_node2

    This function only runs on Windows since the Windows compatible binary is
    provided by the Imperial College team.

    """
    try:
        import imageio
    except ModuleNotFoundError:
        msg = "The imageio package can be installed with pip install imageio"
        raise ModuleNotFoundError(msg)
    file = os.path.splitext(prefix)[0]
    imageio.volsave(file + ".tif", np.array(im.astype("uint8")))
    with open(f"{file}.mhd", "w") as f:
        f.write("ObjectType =  Image\n\
                 NDims =	   3 \n\
                 ElementType = MET_UCHAR \n\
                 DimSize = " + str(im.shape[0]) + " " + str(im.shape[1]) + " " +
                 str(im.shape[2]) + "\n\
                 ElementSpacing =" + str(voxel_size*1e6) + " " +
                 str(voxel_size*1e6) + " " + str(voxel_size*1e6) + "\n\
                 Offset = 0   0  0 \n\
                 ElementDataFile = " + file + ".tif")
    subprocess.Popen([path_to_exe, file + ".mhd"])
    time_elapsed = 0
    while _is_running('pnextract'):
        logger.info('Maximal ball algorithm running for {time_elapsed} s')
        time.sleep(10)
        time_elapsed += 10


def _is_running(process_name):
    r"""
    Check if there is any running process that contains the given name.
    """
    for proc in psutil.process_iter():
        try:
            # Check if process name contains the given name string.
            if process_name.lower() in proc.name().lower():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            pass
    return False



import logging
import numpy as np
import inspect as insp
from edt import edt
import porespy as ps
from numba import njit
import scipy.spatial as sptl
import scipy.ndimage as spim
import scipy.stats as spst
from deprecated import deprecated
from porespy.tools import norm_to_uniform, ps_ball, ps_disk, get_border, ps_round
from porespy.tools import extract_subsection
from porespy.tools import insert_sphere
from porespy.tools import _insert_disk_at_points
from porespy import settings
from typing import List


__all__ = [
    "blobs",
    "bundle_of_tubes",
    "cylinders",
    "cylindrical_plug",
    "insert_shape",
    "lattice_spheres",
    "line_segment",
    "overlapping_spheres",
    "polydisperse_spheres",
    "RSA",
    "rsa",
    "random_spheres",
    "voronoi_edges",
    "_get_Voronoi_edges",
]


tqdm = ps.tools.get_tqdm()
logger = logging.getLogger(__name__)


def cylindrical_plug(shape, r=None, axis=2):
    r"""
    Generates a cylindrical plug suitable for use as a mask on a tomogram

    Parameters
    ----------
    shape : array_like
        The shape of the image to create.  Can be 3D, or 2D in which case
        a circle is returned.
    r : int (optional)
        The diameter of the cylinder to create. If not provided then the
        largest possible radius is used to fit within the image.
    axis : int
        The direction along with the cylinder's axis of rotation should be
        oriented.  The default is 2, which is the z-direction.

    Returns
    -------
    cylinder : ndarray
        A boolean image of the size given by ``shape`` with ``True``'s
        inside the cylinder and ``False``'s outside.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/cylindrical_plug.html>`_
    to view online example.

    """
    shape = np.array(shape, dtype=int)
    axes = np.array(list(set([0, 1, 2]).difference(set([axis]))), dtype=int)
    if len(shape) == 3:
        im2d = np.ones(shape=shape[axes])
        im2d[int(shape[axes[0]]/2), int(shape[axes[1]]/2)] = 0
        dt = edt(im2d)
        if r is None:
            r = int(min(shape[axes])/2)
        circ = dt < r
        tile_ax = [1, 1, 1]
        tile_ax[axis] = shape[axis]
        circ = np.expand_dims(circ, axis)
        cyl = np.tile(circ, tile_ax)
    if len(shape) == 2:
        im2d = np.ones(shape=shape)
        im2d[int(shape[0]/2), int(shape[1]/2)] = 0
        dt = edt(im2d)
        if r is None:
            r = int(min(shape[axes])/2)
        cyl = dt < r
    return cyl


def insert_shape(im, element, center=None, corner=None, value=1, mode="overwrite"):
    r"""
    Inserts sub-image into a larger image at the specified location.

    If the inserted image extends beyond the boundaries of the image it will
    be cropped accordingly.

    Parameters
    ----------
    im : ndarray
        The image into which the sub-image will be inserted
    element : ndarray
        The sub-image to insert
    center : tuple
        Coordinates indicating the position in the main image where the
        inserted imaged will be centered.  If ``center`` is given then
        ``corner`` cannot be specified.  Note that ``center`` can only be
        used if all dimensions of ``element`` are odd, otherwise the meaning
        of center is not defined.
    corner : tuple
        Coordinates indicating the position in the main image where the
        lower corner (i.e. [0, 0, 0]) of the inserted image should be anchored.
        If ``center`` is given then ``corner`` cannot be specified.
    value : scalar
        A scalar value to apply to the sub-image.  The default is 1.
    mode : string
        If 'overwrite' (default) the inserted image replaces the values in the
        main image.  If 'overlay' the inserted image is added to the main
        image.  In both cases the inserted image is multiplied by ``value``
        first.

    Returns
    -------
    im : ndarray
        A copy of ``im`` with the supplied element inserted.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/insert_shape.html>`_
    to view online example.

    """
    im = im.copy()
    if im.ndim != element.ndim:
        msg = f"Image shape {im.shape} and element shape {element.shape} do not match"
        raise Exception(msg)

    s_im = []
    s_el = []
    if (center is not None) and (corner is None):
        for dim in range(im.ndim):
            r, d = np.divmod(element.shape[dim], 2)
            if d == 0:
                raise Exception(
                    "Cannot specify center point when element "
                    + "has one or more even dimension"
                )
            lower_im = np.amax((center[dim] - r, 0))
            upper_im = np.amin((center[dim] + r + 1, im.shape[dim]))
            s_im.append(slice(lower_im, upper_im))
            lower_el = np.amax((lower_im - center[dim] + r, 0))
            upper_el = np.amin((upper_im - center[dim] + r, element.shape[dim]))
            s_el.append(slice(lower_el, upper_el))
    elif (corner is not None) and (center is None):
        for dim in range(im.ndim):
            L = int(element.shape[dim])
            lower_im = np.amax((corner[dim], 0))
            upper_im = np.amin((corner[dim] + L, im.shape[dim]))
            s_im.append(slice(lower_im, upper_im))
            lower_el = np.amax((lower_im - corner[dim], 0))
            upper_el = np.amin((upper_im - corner[dim], element.shape[dim]))
            s_el.append(slice(min(lower_el, upper_el), upper_el))
    else:
        raise Exception("Cannot specify both corner and center")

    if mode == "overlay":
        im[tuple(s_im)] = im[tuple(s_im)] + element[tuple(s_el)] * value
    elif mode == "overwrite":
        im[tuple(s_im)] = element[tuple(s_el)] * value
    else:
        raise Exception("Invalid mode " + mode)
    return im


@deprecated("This function has been renamed to rsa (lowercase to meet pep8)")
def RSA(*args, **kwargs):
    return rsa(*args, **kwargs)


def random_spheres(
    im_or_shape: np.array,
    r: int,
    volume_fraction: int = 1,
    clearance: int = 0,
    protrusion: int = 0,
    n_max: int = 100000,
    edges: str = "contained",
    return_spheres: bool = False,
    smooth: bool = True,
    seed: int = None,
):
    r"""
    Generates a sphere or disk packing using Random Sequential Addition

    Parameters
    ----------
    im_or_shape : ndarray or list
        To provide flexibility, this argument accepts either an image into
        which the spheres are inserted, or a shape which is used to create an
        empty image.  In both cases the spheres are added as ``True`` values
        to the background.  Since ``True`` is considered the pore space, then
        the added spheres represent holes.
    r : int
        The radius of the disk or sphere to insert.
    volume_fraction : scalar (default is 1.0)
        The fraction of the image that should be filled with spheres.  The
        spheres are added as ``True``'s, so each sphere addition increases the
        ``volume_fraction`` until the specified limit is reached.  Note that if
        ``n_max`` is reached first, then ``volume_fraction`` will not be
        achieved.  Also, ``volume_fraction`` is not counted correctly if the
        ``mode`` is ``'extended'``.
    clearance : int (optional, default = 0)
        The amount of space to put between each sphere. Negative values are
        acceptable to create overlaps, so long as ``abs(clearance) < r``.
    protrusion : int (optional, default = 0)
        The amount by which inserted spheres are allowed to protrude outside of
        the given background.  If set to 0 (the default) then all spheres will
        be fully inside the region marked ``False`` in the input image. If > 0, then
        spheres will extend into the region marked ``True`` in the input image.
    n_max : int (default is 100,000)
        The maximum number of spheres to add.  Using a low value may halt
        the addition process prior to reaching the specified
        ``volume_fraction``.  If ``None`` is given, then no limit is used.
    edges : string (default is 'contained')
        Controls how the edges of the image are handled.  Options are:

        ============ ===============================================================
        edges        description
        ============ ===============================================================
        'contained'  Spheres are all completely within the image
        'extended'   Spheres are allowed to extend beyond the edge of the
                     image.  In this mode the volume fraction will be less than
                     requested since some spheres extend beyond the image, but their
                     entire volume is counted as added for computational efficiency.
        ============ ===============================================================

    return_spheres : bool
        If ``True`` then an image containing only the spheres is returned
        rather than the input image with the spheres added, which is the
        default behavior.
    smooth : bool
        Indicates whether balls should have smooth faces (``True``) or should
        include the bumps on the extremities (``False``).
    seed : int
        The seed to supply to the random number generators. Because this function
        uses ``numba`` for speed, calling the normal ``numpy.random.seed(<seed>)``
        has no effect. To get a repeatable image, the seed must be passed to the
        function so it can be initialized the way ``numba`` requires. The default
        is ``None``, which means each call will produce a new realization.

    Returns
    -------
    image : ndarray
        An image with spheres of specified radius *added* to the background.

    See Also
    --------
    pseudo_gravity_packing
    pseudo_electrostatic_packing

    Notes
    -----
    This algorithm ensures that spheres do not overlap but does not
    guarantee they are tightly packed.

    This function adds spheres to the background of the received ``im``, which
    allows iteratively adding spheres of different radii to the unfilled space
    by repeatedly passing in the result of previous calls to RSA.

    References
    ----------
    [1] Random Heterogeneous Materials, S. Torquato (2001)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/random_spheres.html>`_
    to view online example.

    """
    im = rsa(
        im_or_shape=im_or_shape,
        r=r,
        volume_fraction=volume_fraction,
        clearance=clearance,
        protrusion=protrusion,
        n_max=n_max,
        mode=edges,
        return_spheres=return_spheres,
        smooth=smooth,
        seed=seed)
    return im


@deprecated("This function will be renamed random_spheres in a future version")
def rsa(
    im_or_shape: np.array,
    r: int,
    volume_fraction: int = 1,
    clearance: int = 0,
    protrusion: int = 0,
    n_max: int = 100000,
    mode: str = "contained",
    return_spheres: bool = False,
    smooth: bool = True,
    seed: int = None,
):
    r"""
    Generates a sphere or disk packing using Random Sequential Addition

    Parameters
    ----------
    im_or_shape : ndarray or list
        To provide flexibility, this argument accepts either an image into
        which the spheres are inserted, or a shape which is used to create an
        empty image.  In both cases the spheres are added as ``True`` values
        to the background.  Since ``True`` is considered the pore space, then
        the added spheres represent holes.
    r : int
        The radius of the disk or sphere to insert.
    volume_fraction : scalar (default is 1.0)
        The fraction of the image that should be filled with spheres.  The
        spheres are added as ``True``'s, so each sphere addition increases the
        ``volume_fraction`` until the specified limit is reached.  Note that if
        ``n_max`` is reached first, then ``volume_fraction`` will not be
        achieved.  Also, ``volume_fraction`` is not counted correctly if the
        ``mode`` is ``'extended'``.
    clearance : int (optional, default = 0)
        The amount of space to put between each sphere. Negative values are
        acceptable to create overlaps, so long as ``abs(clearance) < r``.
    protrusion : int (optional, default = 0)
        The amount by which inserted spheres are allowed to protrude outside of
        the given background.  If set to 0 (the default) then all spheres will
        be fully inside the region marked ``False`` in the input image. If > 0, then
        spheres will extend into the region marked ``True`` in the input image.
    n_max : int (default is 100,000)
        The maximum number of spheres to add.  Using a low value may halt
        the addition process prior to reaching the specified
        ``volume_fraction``.  If ``None`` is given, then no limit is used.
    mode : string (default is 'contained')
        Controls how the edges of the image are handled.  Options are:

        'contained'
            Spheres are all completely within the image
        'extended'
            Spheres are allowed to extend beyond the edge of the
            image.  In this mode the volume fraction will be less than
            requested since some spheres extend beyond the image, but their
            entire volume is counted as added for computational efficiency.

    return_spheres : bool
        If ``True`` then an image containing only the spheres is returned
        rather than the input image with the spheres added, which is the
        default behavior.
    smooth : bool
        Indicates whether balls should have smooth faces (``True``) or should
        include the bumps on the extremities (``False``).
    seed : int
        The seed to supply to the random number generators. Because this function
        uses ``numba`` for speed, calling the normal ``numpy.random.seed(<seed>)``
        has no effect. To get a repeatable image, the seed must be passed to the
        function so it can be initialized the way ``numba`` requires. The default
        is ``None``, which means each call will produce a new realization.

    Returns
    -------
    image : ndarray
        An image with spheres of specified radius *added* to the background.

    See Also
    --------
    pseudo_gravity_packing
    pseudo_electrostatic_packing

    Notes
    -----
    This algorithm ensures that spheres do not overlap but does not
    guarantee they are tightly packed.

    This function adds spheres to the background of the received ``im``, which
    allows iteratively adding spheres of different radii to the unfilled space
    by repeatedly passing in the result of previous calls to RSA.

    References
    ----------
    [1] Random Heterogeneous Materials, S. Torquato (2001)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/rsa.html>`_
    to view online example.

    """
    logger.debug(f"rsa: Adding spheres of size {r}")
    if np.array(im_or_shape).ndim < 2:
        im = np.zeros(shape=im_or_shape, dtype=bool)
        input_im = np.copy(im)
    else:
        input_im = np.copy(im_or_shape)
        im = np.zeros_like(im_or_shape, dtype=bool)
    if seed is not None:  # Initialize rng so numba sees it
        _set_seed(seed)
    im = im.astype(bool)
    shape_orig = im.shape  # Store original image shape, to undo padding at the end
    if n_max is None:
        n_max = np.inf
    # Compute volume fraction info
    vf_final = volume_fraction
    vf_start = im.sum(dtype=np.int64) / im.size
    vf_template = ps_round(r, ndim=im.ndim, smooth=smooth).sum(dtype=np.int64) / im.size
    logger.debug(f"Initial volume fraction: {vf_start}")
    # Dilate existing objects by strel to remove pixels near them
    # from consideration for sphere placement
    logger.info("Dilating foreground features by sphere radius")
    dt = edt(input_im == 0)
    options_im = dt >= (r - protrusion)
    # Depending on mode, adjust options_im to remove options around edge
    if mode == "contained":
        border = get_border(im.shape, thickness=r, mode="faces")
        options_im[border] = False
    elif mode == "extended":
        im = np.pad(im, pad_width=r, mode="edge")
        options_im = np.pad(options_im, r, mode='symmetric')
    else:
        raise Exception("Unrecognized mode: ", mode)
    # Begin inserting the spheres
    vf = vf_start
    free_sites = np.flatnonzero(options_im)
    i = 0
    while (vf <= vf_final) and (i < n_max) and (len(free_sites) > 0):
        # Choose a random site from free_sites
        c, count = _make_choice(options_im, free_sites=free_sites)
        # The 100 below is arbitrary and may change performance
        if count > 100:
            # Regenerate list of free_sites
            logger.debug(f"Regenerating `free_sites` after {i} iterations")
            free_sites = np.flatnonzero(options_im)
        if all(np.array(c) == -1):
            break
        im = _insert_disk_at_points(im=im,
                                    coords=np.vstack(c),
                                    r=r,
                                    v=True,
                                    smooth=smooth)
        options_im = _insert_disk_at_points(im=options_im,
                                            coords=np.vstack(c),
                                            r=2*r + int(np.round(clearance/2)),
                                            v=False,
                                            smooth=smooth,
                                            overwrite=True)
        vf += vf_template
        i += 1
    logger.info(f"Number of spheres inserted: {i}")
    im = extract_subsection(im, shape_orig)
    logger.debug("Final volume fraction:", vf)
    if not return_spheres:
        im = im + input_im
    return im


@njit
def _set_seed(a):
    np.random.seed(a)


@njit
def _get_rand_float(*args):
    return np.random.rand(*args)


@njit
def _get_rand_int(*args):
    return np.random.randint(*args)


@njit
def _make_choice(options_im, free_sites):
    r"""
    This function is called by _begin_inserting to find valid insertion
    points.

    Parameters
    ----------
    options_im : ndarray
        An array with ``True`` at all valid locations and ``False`` at all
        locations where a sphere already exists PLUS a region of radius R
        around each sphere since these points are also invalid insertion
        points.
    free_sites : array_like
        A 1D array containing valid insertion indices.  This list is used
        to select insertion points from a limited which occasionally gets
        smaller.

    Returns
    -------
    coords : list
        The XY or XYZ coordinates of the next insertion point
    count : int
        The number of attempts that were needed to find valid point.  If
        this value gets too high, a short list of ``free_sites`` should be
        generated in the calling function.

    """
    choice = False
    count = 0
    upper_limit = len(free_sites)
    maxiter = upper_limit * 20
    if options_im.ndim == 2:
        coords = [-1, -1]
        Nx, Ny = options_im.shape
        while not choice:
            if count >= maxiter:
                coords = [-1, -1]
                break
            ind = _get_rand_int(0, upper_limit)
            # This numpy function is not supported by numba yet
            # c1, c2 = np.unravel_index(free_sites[ind], options_im.shape)
            # So using manual unraveling
            coords[1] = free_sites[ind] % Ny
            coords[0] = (free_sites[ind] // Ny) % Nx
            choice = options_im[coords[0], coords[1]]
            count += 1
    if options_im.ndim == 3:
        coords = [-1, -1, -1]
        Nx, Ny, Nz = options_im.shape
        while not choice:
            if count >= maxiter:
                coords = [-1, -1, -1]
                break
            ind = _get_rand_int(0, upper_limit)
            # This numpy function is not supported by numba yet
            # c1, c2, c3 = np.unravel_index(free_sites[ind], options_im.shape)
            # So using manual unraveling
            coords[2] = free_sites[ind] % Nz
            coords[1] = (free_sites[ind] // Nz) % Ny
            coords[0] = (free_sites[ind] // (Nz * Ny)) % Nx
            choice = options_im[coords[0], coords[1], coords[2]]
            count += 1
    return coords, count


def bundle_of_tubes(
    shape: List[int],
    spacing: int,
    distribution=None,
    smooth=True,
    seed=None,
):
    r"""
    Create a 3D image of a bundle of tubes, in the form of a rectangular
    plate with randomly sized holes through it.

    Parameters
    ----------
    shape : list
        The size the image, with the 3rd dimension indicating the plate
        thickness.  If the 3rd dimension is not given then a thickness of
        1 voxel is assumed.
    spacing : int
        The center to center distance of the holes.  The hole sizes will
        be distributed between this values down to 3 voxels.
    distribution : scipy.stats object
        A handle to a scipy stats object with the desired parameters.
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/bundle_of_tubes.html>`_
    to view online example.

    """
    if seed is not None:
        np.random.seed(seed)
    shape = np.array(shape)
    if len(shape) == 2:
        shape = np.hstack((shape, [1]))
    shape2 = shape[shape > 1]
    im = ~lattice_spheres(shape=shape2,
                          r=1,
                          offset=0.5*spacing,
                          spacing=spacing,
                          lattice='sc')
    N = im.sum(dtype=np.int64)
    if distribution is None:
        # +1 below is because randint 4.X gives a max of 3
        distribution = spst.randint(low=3, high=int(spacing/2 + 1))
        Rs = distribution.rvs(N)
    else:
        Rs = distribution.rvs(N)
        Rs = np.around(np.clip(Rs, a_min=1, a_max=spacing/2), decimals=0).astype(int)
    temp = np.zeros_like(im)
    inds = np.where(im)
    for i in range(len(inds[0])):
        c = np.hstack([j[i] for j in inds])
        temp = insert_sphere(im=temp, c=c, r=Rs[i])
    # Add 3rd dimension back
    temp = np.tile(np.atleast_3d(temp), [1, 1, shape[2]])
    return temp


def polydisperse_spheres(shape: List[int],
                         porosity: float,
                         dist,
                         nbins: int = 5,
                         r_min: int = 5,
                         seed=None):
    r"""
    Create an image of randomly placed, overlapping spheres with a
    distribution of radii.

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where Ni is the
        number of voxels in each direction.  If shape is only 2D, then an
        image of polydisperse disks is returns
    porosity : float
        The porosity of the image, defined as the number of void voxels
        divided by the number of voxels in the image. The specified value
        is only matched approximately, so it's suggested to check this
        value after the image is generated.
    dist : scipy.stats distribution object
        This should be an initialized distribution chosen from the large
        number of options in the ``scipy.stats`` submodule.  For instance,
        a normal distribution with a mean of 20 and a standard deviation
        of 10 can be obtained with
        ``dist = scipy.stats.norm(loc=20, scale=10)``
    nbins : int
        The number of discrete sphere sizes that will be used to generate
        the image. This function generates ``nbins`` images of
        monodisperse spheres that span 0.05 and 0.95 of the possible
        values produced by the provided distribution, then overlays them
        to get polydispersivity.
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/polydisperse_spheres.html>`_
    to view online example.

    """
    if seed is not None:
        np.random.seed(seed)
    shape = np.array(shape)
    if np.size(shape) == 1:
        shape = np.full((3,), int(shape))
    Rs = dist.interval(np.linspace(0.05, 0.95, nbins))
    Rs = np.vstack(Rs).T
    Rs = (Rs[:-1] + Rs[1:]) / 2
    Rs = np.clip(Rs.flatten(), a_min=r_min, a_max=None)
    phi_desired = 1 - (1 - porosity) / (len(Rs))
    im = np.ones(shape, dtype=bool)
    for r in Rs:
        phi_im = im.sum(dtype=np.int64) / np.prod(shape)
        phi_corrected = 1 - (1 - phi_desired) / phi_im
        temp = overlapping_spheres(shape=shape, r=r, porosity=phi_corrected)
        im = im * temp
    return im


def voronoi_edges(
    shape: List[int],
    ncells: int,
    r: int = 0,
    flat_faces: bool = True,
    seed=None,
    **kwargs,
):
    r"""
    Create an image from the edges of a Voronoi tessellation.

    Parameters
    ----------
    shape : array_like
        The size of the image to generate in [Nx, Ny, <Nz>] where Ni is the
        number of voxels in each direction.  If Nz is not given a 2D image
        is returned.
    ncells : int
        The number of Voronoi cells to include in the tesselation.
    r : int, optional
        The radius to which Voronoi edges should be dilated in the final
        image.  If not given then edges are a single pixel/voxel thick.
    flat_faces : bool
        Whether the Voronoi edges should lie on the boundary of the
        image (``True``), or if edges outside the image should be removed
        (``False``).
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/voronoi_edges.html>`_
    to view online example.

    """
    if 'radius' in kwargs.keys():
        r = kwargs['radius']
        print('radius keyword is deprecated in favor of just r')
    if seed is not None:
        np.random.seed(seed)
    logger.info(f"Generating {ncells} cells")
    shape = np.array(shape)
    if np.size(shape) == 1:
        shape = np.full((3,), int(shape))
    im = np.zeros(shape, dtype=bool)
    base_pts = tuple(np.around(np.random.rand(ncells, im.ndim) * (shape-1),
                               decimals=0).T.astype(int))
    im[tuple(base_pts)] = True
    pw = [(s, s) for s in im.shape]
    if flat_faces:
        im = np.pad(im, pad_width=pw, mode='symmetric')
    else:
        im = np.pad(im, pad_width=pw, mode='constant', constant_values=0)
    base_pts = np.where(im)
    vor = sptl.Voronoi(points=np.vstack(base_pts).T)
    vor.vertices = np.around(vor.vertices)
    vor.vertices *= (np.array(im.shape) - 1) / np.array(im.shape)
    vor.edges = _get_Voronoi_edges(vor)
    im = np.zeros_like(im)
    for row in vor.edges:
        pts = np.around(vor.vertices[row], decimals=0).astype(int)
        if np.all(pts >= 0) and np.all(pts < im.shape):
            line_pts = line_segment(pts[0], pts[1])
            im[tuple(line_pts)] = True
    im = extract_subsection(im=im, shape=shape)
    im = edt(~im) > r
    return im


def _get_Voronoi_edges(vor):
    r"""
    Given a Voronoi object as produced by the scipy.spatial.Voronoi class,
    this function calculates the start and end points of each edge in the
    Voronoi diagram, in terms of the vertex indices used by the received
    Voronoi object.

    Parameters
    ----------
    vor : scipy.spatial.Voronoi object

    Returns
    -------
    A 2-by-N array of vertex indices, indicating the start and end points
    of each vertex in the Voronoi diagram.  These vertex indices can be
    used to index straight into the ``vor.vertices`` array to get spatial
    positions.

    """
    edges = [[], []]
    for facet in vor.ridge_vertices:
        # Create a closed cycle of vertices that define the facet
        edges[0].extend(facet[:-1] + [facet[-1]])
        edges[1].extend(facet[1:] + [facet[0]])
    edges = np.vstack(edges).T  # Convert to scipy-friendly format
    mask = np.any(edges == -1, axis=1)  # Identify edges at infinity
    edges = edges[~mask]  # Remove edges at infinity
    edges = np.sort(edges, axis=1)  # Move all points to upper triangle
    # Remove duplicate pairs
    edges = edges[:, 0] + 1j * edges[:, 1]  # Convert to imaginary
    edges = np.unique(edges)  # Remove duplicates
    edges = np.vstack((np.real(edges), np.imag(edges))).T  # Back to real
    edges = np.array(edges, dtype=int)
    return edges


def lattice_spheres(shape: List[int],
                    r: int,
                    spacing: int = None,
                    offset: int = None,
                    smooth: bool = True,
                    lattice: str = "sc"):
    r"""
    Generate a cubic packing of spheres in a specified lattice arrangement.

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where N is
        the number of voxels in each direction. For a 2D image, use
        [Nx, Ny].
    radius : int
        The radius of spheres (circles) in the packing.
    spacing : int or List[int]
        The spacing between unit cells. If the spacing is too small then
        spheres may overlap. If an ``int`` is given it will be applied
        in all directions, while a list of ``int`` will be interpreted
        to apply along each axis.
    offset : int or List[int]
        The amount offset to add between sphere centers and the edges of
        the image. A single ``int`` will be applied in all directions,
        while a list of ``int`` will be interpreted to apply along each
        axis.
    smooth : bool, default=True
        If ``True`` (default) the outer extremities of the sphere will
        not have the little bumps on each face.
    lattice : str
        Specifies the type of lattice to create. Options are:

        ======= ===============================================================
        option  description
        ======= ===============================================================
        'sc'    Simple cubic (default), works in both 2D and 3D.
        'tri'   Triangular, only works in 2D
        'fcc'   Face centered cubic, only works in 3D
        'bcc'   Body centered cubic, only works on 3D
        ======= ===============================================================

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space.

    Notes
    -----
    For 2D images, 'sc' gives a square lattice and both 'fcc' and
    'bcc' give a triangular lattice.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/lattice_spheres.html>`_
    to view online example.

    """
    logger.debug(f"Generating {lattice} lattice")
    shape = np.array(shape)
    im = np.zeros(shape, dtype=bool)

    # Parse lattice type
    lattice = lattice.lower()
    if im.ndim == 2:
        if lattice in ['sc', 'square', 'cubic', 'simple cubic']:
            lattice = 'sq'
        elif lattice in ['tri', 'triangular']:
            lattice = 'tri'
        else:
            raise Exception(f'Unrecognized mode: {lattice}')
    else:
        if lattice in ['sc', 'cubic', 'simple cubic']:
            lattice = 'sc'
        elif lattice in ['bcc', 'body centered cubic']:
            lattice = 'bcc'
        elif lattice in ['fcc', 'face centered cubic']:
            lattice = 'fcc'
        else:
            raise Exception(f'Unrecognized mode: {lattice}')

    # Parse offset and spacing args
    if spacing is None:
        spacing = 2*r
    if isinstance(spacing, (int, float)):
        spacing = int(spacing)
        spacing = [spacing]*im.ndim
    spacing = np.array(spacing, dtype=int)
    if offset is None:
        offset = r
    if isinstance(offset, (int, float)):
        offset = int(offset)
        offset = [offset]*im.ndim
    offset = np.array(offset, dtype=int)

    if lattice == 'sq':
        im[offset[0]::spacing[0],
           offset[1]::spacing[1]] = True
    elif lattice == 'tri':
        im[offset[0]::spacing[0],
           offset[1]::spacing[1]] = True
        im[offset[0]+int(spacing[0]/2)::spacing[0],
           offset[1]+int(spacing[1]/2)::spacing[1]] = True
    elif lattice == 'sc':
        im[offset[0]::spacing[0],
           offset[1]::spacing[1],
           offset[2]::spacing[2]] = True
    elif lattice == 'bcc':
        im[offset[0]::spacing[0],
           offset[1]::spacing[1],
           offset[2]::spacing[2]] = True
        im[offset[0]+int(spacing[0]/2)::spacing[0],
           offset[1]+int(spacing[1]/2)::spacing[1],
           offset[2]+int(spacing[2]/2)::spacing[2]] = True
    elif lattice == 'fcc':
        im[offset[0]::spacing[0],
           offset[1]::spacing[1],
           offset[2]::spacing[2]] = True
        # xy-plane
        im[offset[0]+int(spacing[0]/2)::spacing[0],
           offset[1]+int(spacing[1]/2)::spacing[1],
           offset[2]::spacing[2]] = True
        # xz-plane
        im[offset[0]+int(spacing[0]/2)::spacing[0],
           offset[1]::spacing[1],
           offset[2]+int(spacing[2]/2)::spacing[2]] = True
        # yz-plane
        im[offset[0]::spacing[0],
           offset[1]+int(spacing[1]/2)::spacing[1],
           offset[2]+int(spacing[2]/2)::spacing[2]] = True
    if smooth:
        im = ~(edt(~im) < r)
    else:
        im = ~(edt(~im) <= r)
    return im


def overlapping_spheres(
    shape: List[int],
    r: int,
    porosity: float,
    maxiter: int = 10,
    tol: float = 0.01,
    seed=None,
):
    r"""
    Generate a packing of overlapping mono-disperse spheres

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where Ni is the
        number of voxels in the i-th direction.
    r : scalar
        The radius of spheres in the packing.
    porosity : scalar
        The porosity of the final image, accurate to the given tolerance.
    maxiter : int
        Maximum number of iterations for the iterative algorithm that improves
        the porosity of the final image to match the given value.
    tol : float
        Tolerance for porosity of the final image compared to the given value.
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space

    Notes
    -----
    This method can also be used to generate a dispersion of hollows by
    treating ``porosity`` as solid volume fraction and inverting the
    returned image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/overlapping_spheres.html>`_
    to view online example.

    """
    if seed is not None:
        np.random.seed(seed)
    shape = np.array(shape)
    if np.size(shape) == 1:
        shape = np.full((3, ), int(shape))
    ndim = (shape != 1).sum(dtype=np.int64)
    s_vol = ps_disk(r).sum(dtype=np.int64) if ndim == 2 \
        else ps_ball(r).sum(dtype=np.int64)

    bulk_vol = np.prod(shape)
    N = int(np.ceil((1 - porosity) * bulk_vol / s_vol))
    im = np.random.random(size=shape)

    # Helper functions for calculating porosity: phi = g(f(N))
    def f(N):
        return edt(im > N / bulk_vol) < r

    def g(im):
        r"""Returns fraction of 0s, given a binary image"""
        return 1 - im.sum(dtype=np.int64) / np.prod(shape)

    # # Newton's method for getting image porosity match the given
    # w = 1.0                         # Damping factor
    # dN = 5 if ndim == 2 else 25     # Perturbation
    # for i in range(maxiter):
    #     err = g(f(N)) - porosity
    #     d_err = (g(f(N+dN)) - g(f(N))) / dN
    #     if d_err == 0:
    #         break
    #     if abs(err) <= tol:
    #         break
    #     N2 = N - int(err/d_err)   # xnew = xold - f/df
    #     N = w * N2 + (1-w) * N

    # Bisection search: N is always undershoot (bc. of overlaps)
    N_low, N_high = N, 4 * N
    for i in range(maxiter):
        N = np.mean([N_high, N_low], dtype=int)
        err = g(f(N)) - porosity
        if err > 0:
            N_low = N
        else:
            N_high = N
        if abs(err) <= tol:
            break

    return ~f(N)


def blobs(
    shape: List[int],
    porosity: float = 0.5,
    blobiness: int = 1,
    divs: int = 1,
    seed=None,
):
    """
    Generates an image containing amorphous blobs

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where N is the
        number of voxels
    porosity : float
        If specified, this will threshold the image to the specified value
        prior to returning.  If ``None`` is specified, then the scalar
        noise field is converted to a uniform distribution and returned
        without thresholding.
    blobiness : int or list of ints(default = 1)
        Controls the morphology of the blobs.  A higher number results in
        a larger number of small blobs.  If a list is supplied then the
        blobs are anisotropic.
    divs : int or array_like
        The number of times to divide the image for parallel processing.
        If ``1`` then parallel processing does not occur.  ``2`` is
        equivalent to ``[2, 2, 2]`` for a 3D image.  The number of cores
        used is specified in ``porespy.settings.ncores`` and defaults to
        all cores.
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space

    See Also
    --------
    norm_to_uniform

    Notes
    -----
    This function generates random noise, the applies a gaussian blur to
    the noise with a sigma controlled by the blobiness argument as:

        $$ np.mean(shape) / (40 * blobiness) $$

    The value of 40 was chosen so that a ``blobiness`` of 1 gave a
    reasonable result.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/blobs.html>`_
    to view online example.

    """
    if seed is not None:
        np.random.seed(seed)
    if isinstance(shape, int):
        shape = [shape]*3
    if len(shape) == 1:
        shape = [shape[0]]*3
    shape = np.array(shape)
    if isinstance(blobiness, int):
        blobiness = [blobiness]*len(shape)
    blobiness = np.array(blobiness)
    parallel = False
    if isinstance(divs, int):
        divs = [divs]*len(shape)
    if max(divs) > 1:
        parallel = True
        logger.info(f'Performing {insp.currentframe().f_code.co_name} in parallel')
    sigma = np.mean(shape) / (40 * blobiness)
    im = np.random.random(shape)
    if parallel:
        overlap = max([int(s*4) for s in np.array(sigma, ndmin=1)])
        im = ps.filters.chunked_func(func=spim.gaussian_filter,
                                     input=im, sigma=sigma,
                                     divs=divs, overlap=overlap)
    else:
        im = spim.gaussian_filter(im, sigma=sigma)
    im = norm_to_uniform(im, scale=[0, 1])
    if porosity:
        im = im < porosity
    return im


def _cylinders(
    shape: List[int],
    r: int,
    ncylinders: int,
    phi_max: float = 0,
    theta_max: float = 90,
    length: float = None,
    verbose: bool = True,
    seed=None,
):
    r"""
    Generates a binary image of overlapping cylinders.

    This is a good approximation of a fibrous mat.

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where N is the
        number of voxels. 2D images are not permitted.
    r : int
        The radius of the cylinders in voxels
    ncylinders : int
        The number of cylinders to add to the domain. Adjust this value to
        control the final porosity, which is not easily specified since
        cylinders overlap and intersect different fractions of the domain.
    phi_max : int
        A value between 0 and 90 that controls the amount that the
        cylinders lie *out of* the XY plane, with 0 meaning all cylinders
        lie in the XY plane, and 90 meaning that cylinders are randomly
        oriented out of the plane by as much as +/- 90 degrees.
    theta_max : int
        A value between 0 and 90 that controls the amount of rotation
        *in the* XY plane, with 0 meaning all cylinders point in the
        X-direction, and 90 meaning they are randomly rotated about the Z
        axis by as much as +/- 90 degrees.
    length : int
        The length of the cylinders to add.  If ``None`` (default) then
        the cylinders will extend beyond the domain in both directions so
        no ends will exist. If a scalar value is given it will be
        interpreted as the Euclidean distance between the two ends of the
        cylinder. Note that one or both of the ends *may* still lie
        outside the domain, depending on the randomly chosen center point
        of the cylinder.
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space

    """
    if seed is not None:
        np.random.seed(seed)
    shape = np.array(shape)
    if np.size(shape) == 1:
        shape = np.full((3, ), int(shape))
    elif np.size(shape) == 2:
        raise Exception("2D cylinders don't make sense")
    # Find hypotenuse of domain from [0,0,0] to [Nx,Ny,Nz]
    H = np.sqrt(np.sum(np.square(shape), dtype=np.int64)).astype(int)
    if length is None:  # Assume cylinders span domain if length not given
        length = 2 * H
    R = min(int(length / 2), 2 * H)  # Trim given length to 2H if too long
    # Adjust max angles to be between 0 and 90
    if (phi_max > 90) or (phi_max < 0):
        raise Exception('phi_max must be betwen 0 and 90')
    if (theta_max > 90) or (theta_max < 0):
        raise Exception('theta_max must be betwen 0 and 90')
    # Create empty image for inserting into
    im = np.zeros(shape, dtype=bool)
    n = 0
    L = min(H, R)
    # Disable tqdm if called from another tqdm to prevent double pbars
    tqdm_settings = settings.tqdm.copy()
    if not settings.tqdm["disable"]:
        tqdm_settings = {**settings.tqdm, **{'disable': not verbose}}
    with tqdm(ncylinders, **tqdm_settings) as pbar:
        while n < ncylinders:
            # Choose a random starting point in domain
            x = np.random.rand(3) * (shape + 2 * L)
            # Chose a random phi and theta within given ranges
            phi = (np.pi / 2 - np.pi * np.random.rand()) * phi_max / 90
            theta = (np.pi / 2 - np.pi * np.random.rand()) * theta_max / 90
            X0 = R * np.array([np.cos(phi) * np.cos(theta),
                               np.cos(phi) * np.sin(theta),
                               np.sin(phi)])
            [X0, X1] = [x + X0, x - X0]
            crds = line_segment(X0, X1)
            lower = ~np.any(np.vstack(crds).T < [L, L, L], axis=1)
            upper = ~np.any(np.vstack(crds).T >= shape + L, axis=1)
            valid = upper * lower
            if np.any(valid):
                im[crds[0][valid] - L, crds[1][valid] - L, crds[2][valid] - L] = 1
                n += 1
                pbar.update()
    im = np.array(im, dtype=bool)
    dt = edt(~im) < r
    return ~dt


def cylinders(
    shape: List[int],
    r: int,
    ncylinders: int = None,
    porosity: float = None,
    phi_max: float = 0,
    theta_max: float = 90,
    length: float = None,
    maxiter: int = 3,
    seed=None,
):
    r"""
    Generates a binary image of overlapping cylinders given porosity OR
    number of cylinders.

    This is a good approximation of a fibrous mat.

    Parameters
    ----------
    shape : list
        The size of the image to generate in [Nx, Ny, Nz] where N is the
        number of voxels. 2D images are not permitted.
    r : scalar
        The radius of the cylinders in voxels
    ncylinders : int
        The number of cylinders to add to the domain. Adjust this value to
        control the final porosity, which is not easily specified since
        cylinders overlap and intersect different fractions of the domain.
    porosity : float
        The targeted value for the porosity of the generated mat. The
        function uses an algorithm for predicted the number of required
        number of cylinder, and refines this over a certain number of
        fractional insertions (according to the 'iterations' input).
    phi_max : int
        A value between 0 and 90 that controls the amount that the
        cylinders lie *out of* the XY plane, with 0 meaning all cylinders
        lie in the XY plane, and 90 meaning that cylinders are randomly
        oriented out of the plane by as much as +/- 90 degrees.
    theta_max : int
        A value between 0 and 90 that controls the amount of rotation
        *in the* XY plane, with 0 meaning all cylinders point in the
        X-direction, and 90 meaning they are randomly rotated about the Z
        axis by as much as +/- 90 degrees.
    length : int
        The length of the cylinders to add.  If ``None`` (default) then
        the cylinders will extend beyond the domain in both directions so
        no ends will exist. If a scalar value is given it will be
        interpreted as the Euclidean distance between the two ends of the
        cylinder. Note that one or both of the ends *may* still lie
        outside the domain, depending on the randomly chosen center point
        of the cylinder.
    maxiter : int
        The number of fractional fiber insertions used to target the
        requested porosity. By default a value of 3 is used (and this is
        typically effective in getting very close to the targeted
        porosity), but a greater number can be input to improve the
        achieved porosity.
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    image : ndarray
        A boolean array with ``True`` values denoting the pore space

    Notes
    -----
    The cylinders_porosity function works by estimating the number of
    cylinders needed to be inserted into the domain by estimating
    cylinder length, and exploiting the fact that, when inserting any
    potentially overlapping objects randomly into a volume v_total (which
    has units of pixels and is equal to dimx x dimy x dimz, for example),
    such that the total volume of objects added to the volume is v_added
    (and includes any volume that was inserted but overlapped with already
    occupied space), the resulting porosity will be equal to
    exp(-v_added/v_total).

    After intially estimating the cylinder number and inserting a small
    fraction of the estimated number, the true cylinder volume is
    calculated, the estimate refined, and a larger fraction of cylinders
    inserted. This is repeated a number of times according to the
    ``maxiter`` argument, yielding an image with a porosity close to
    the goal.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/cylinders.html>`_
    to view online example.

    """
    if seed is not None:
        np.random.seed(seed)

    if ncylinders is not None:
        im = _cylinders(
            shape=shape,
            r=r,
            ncylinders=ncylinders,
            phi_max=phi_max,
            theta_max=theta_max,
            length=length,
        )
        return im

    if porosity is None:
        raise Exception("'ncylinders' and 'porosity' can't be both None")

    # if maxiter < 3:
    #     raise Exception("Iterations must be greater than or equal to 3")

    vol_total = float(np.prod(shape))

    def get_num_pixels(porosity):
        r"""
        Helper method to calculate number of pixels given a porosity
        """
        return -np.log(porosity) * vol_total

    # Crudely estimate fiber length as cube root of product of dims
    length_estimate = vol_total ** (1 / 3) if length is None else length

    # Rough fiber volume estimate
    vol_fiber = length_estimate * np.pi * r * r
    n_pixels_to_add = get_num_pixels(porosity)

    # Rough estimate of n_fibers
    n_fibers_added = 0
    # Calculate fraction of fibers to be added in each iteration.
    subdif = 0.8 / np.sum(np.arange(1, maxiter) ** 2, dtype=np.int64)
    fractions = [0.2]
    for i in range(1, maxiter):
        fractions.append(fractions[i - 1] + (maxiter - i) ** 2 * subdif)

    im = np.ones(shape, dtype=bool)
    for frac in tqdm(fractions, **settings.tqdm):
        n_fibers_total = n_pixels_to_add / vol_fiber
        n_fibers = int(np.ceil(frac * n_fibers_total) - n_fibers_added)
        if n_fibers > 0:
            im = im & _cylinders(shape, r, n_fibers,
                                 phi_max, theta_max, length,
                                 verbose=False)
        n_fibers_added += n_fibers
        # Update parameters for next iteration
        porosity = ps.metrics.porosity(im)
        vol_added = get_num_pixels(porosity)
        vol_fiber = vol_added / n_fibers_added

    logger.debug(f"{n_fibers_added} fibers added to reach target porosity.")

    return im


def line_segment(X0, X1):
    r"""
    Calculate the voxel coordinates of a straight line between the two
    given end points

    Parameters
    ----------
    X0 and X1 : array_like
        The [x, y] or [x, y, z] coordinates of the start and end points of
        the line.

    Returns
    -------
    coords : list of lists
        A list of lists containing the X, Y, and Z coordinates of all
        voxels that should be drawn between the start and end points to
        create a solid line.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/line_segment.html>`_
    to view online example.

    """
    X0 = np.around(X0).astype(int)
    X1 = np.around(X1).astype(int)
    if len(X0) == 3:
        L = np.amax(
            np.absolute([[X1[0] - X0[0]], [X1[1] - X0[1]], [X1[2] - X0[2]]])
        ) + 1
        x = np.rint(np.linspace(X0[0], X1[0], L)).astype(int)
        y = np.rint(np.linspace(X0[1], X1[1], L)).astype(int)
        z = np.rint(np.linspace(X0[2], X1[2], L)).astype(int)
        return [x, y, z]
    else:
        L = np.amax(np.absolute([[X1[0] - X0[0]], [X1[1] - X0[1]]])) + 1
        x = np.rint(np.linspace(X0[0], X1[0], L)).astype(int)
        y = np.rint(np.linspace(X0[1], X1[1], L)).astype(int)
        return [x, y]



import logging
import numpy as np
import scipy.ndimage as spim
from porespy.tools import get_tqdm
from porespy import settings


tqdm = get_tqdm()
logger = logging.getLogger(__name__)


__all__ = [
    'random_cantor_dust',
    'sierpinski_foam',
    'sierpinski_foam2',
]


def random_cantor_dust(shape, n=5, p=2, f=0.8, seed=None):
    r"""
    Generates an image of random cantor dust

    Parameters
    ----------
    shape : array_like
        The shape of the final image.  If not evenly divisible by $p**n$
        it will be increased to the nearest size that is.
    n : int
        The number of times to iteratively divide the image.
    p : int (default = 2)
        The number of divisions to make on each iteration.
    f : float (default = 0.8)
        The fraction of the set to keep on each iteration.
    seed : int, optional, default = `None`
        Initializes numpy's random number generator to the specified state. If not
        provided, the current global value is used. This means calls to
        ``np.random.state(seed)`` prior to calling this function will be respected.

    Returns
    -------
    dust : ndarray
        A boolean image of a random Cantor dust

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/randon_cantor_dust.html>`_
    to view online example.

    """
    if seed is not None:
        np.random.seed(seed)
    # Parse the given shape and adjust if necessary
    shape = np.array(shape)
    trim = np.mod(shape, (p**n))
    if np.any(trim > 0):
        shape = shape - trim + p**n
        logger.warning(f"Requested shape being changed to {shape}")
    im = np.ones(shape, dtype=bool)
    divs = []
    if isinstance(n, int):
        for i in range(1, n):
            divs.append(p**i)
    else:
        for i in n:
            divs.append(p**i)
    for i in tqdm(divs, **settings.tqdm):
        sh = (np.array(im.shape)/i).astype(int)
        mask = np.random.rand(*sh) < f
        mask = spim.zoom(mask, zoom=i, order=0)
        im = im*mask
    return im


def sierpinski_foam2(shape, n=5):
    r"""
    Generates an image of a Sierpinski carpet or foam with independent control of
    image size and number of iterations

    Parameters
    ----------
    shape : array_like
        The shape of the final image to create. To create a 'centered' image,
        the shape should be ``3**n``.
    n : int
        The number of times to iteratively divide the image. This functions starts
        by inserting single voxels, then inserts increasingly large squares/cubes.

    Returns
    -------
    im : ndarray
        A boolean image with ``False`` values inserted at at the center of each
        square (or cubic) sub-section.

    Notes
    -----
    This function may generate a larger image than need then return the center
    portion of the requested ``shape``, so the edges may be clipped from the
    true Sierpinski foam. This can be avoided by setting shape to some multiple
    of ``3**n``.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/sierpinski_foam2.html>`_
    to view online example.

    """
    im = np.zeros(shape, dtype=bool)
    if im.ndim == 2:
        im[1::3, 1::3] = 1
    else:
        im[1::3, 1::3, 1::3] = 1
    i = 1
    pbar = tqdm()
    while i < n:
        if im.ndim == 2:
            mask = np.zeros([3**(i+1), 3**(i+1)], dtype=bool)
            s = 3**(i+1)//3
            mask[s:-s, s:-s] = 1
            t = int(np.ceil(im.shape[0]/mask.shape[0]))
            im2 = np.tile(mask, [t, t])
            im2 = im2[:im.shape[0], :im.shape[1]]
        if im.ndim == 3:
            mask = np.zeros([3**(i+1), 3**(i+1), 3**(i+1)], dtype=bool)
            s = 3**(i+1)//3
            mask[s:-s, s:-s, s:-s] = 1
            t = int(np.ceil(im.shape[0]/mask.shape[0]))
            im2 = np.tile(mask, [t, t, t])
            im2 = im2[:im.shape[0], :im.shape[1], :im.shape[2]]
        im += im2
        i += 1
        pbar.update()
    pbar.close()
    im = im == 0
    return im


def sierpinski_foam(dmin=1, n=5, ndim=2, max_size=1e9):
    r"""
    Generates an image of a Sierpinski carpet or foam

    Parameters
    ----------
    dmin : int
        The size of the smallest square in the final image
    n : int
        The number of times to iteratively tile the image
    ndim : int
        The number of dimensions of the desired image, can be 2 or 3. The
        default value is 2.

    Returns
    -------
    foam : ndarray
        A boolean image of a Sierpinski gasket or foam

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/sierpinski_foam.html>`_
    to view online example.

    """
    def _insert_cubes(im, n):
        if n > 0:
            n -= 1
            shape = np.asarray(np.shape(im))
            im = np.tile(im, (3, 3, 3))
            im[shape[0]:2*shape[0], shape[1]:2*shape[1], shape[2]:2*shape[2]] = 0
            if im.size < max_size:
                im = _insert_cubes(im, n)
        return im

    def _insert_squares(im, n):
        if n > 0:
            n -= 1
            shape = np.asarray(np.shape(im))
            im = np.tile(im, (3, 3))
            im[shape[0]:2*shape[0], shape[1]:2*shape[1]] = 0
            if im.size < max_size:
                im = _insert_squares(im, n)
        return im

    im = np.ones([dmin]*ndim, dtype=int)
    if ndim == 2:
        im = _insert_squares(im, n)
    elif ndim == 3:
        im = _insert_cubes(im, n)
    return im



import os
import subprocess
import numpy as np
import scipy.ndimage as nd
import skimage.measure as ms
from porespy.tools import sanitize_filename
from porespy.networks import generate_voxel_image
from porespy.filters import reduce_peaks
from skimage.morphology import ball
from edt import edt


def dict_to_vtk(data, filename, voxel_size=1, origin=(0, 0, 0)):
    r"""
    Accepts multiple images as a dictionary and compiles them into a vtk file

    Parameters
    ----------
    data : dict
        A dictionary of *key: value* pairs, where the *key* is the name of the
        scalar property stored in each voxel of the array stored in the
        corresponding *value*.
    path : string
        Path to output file
    voxel_size : int
        The side length of the voxels (voxels  are cubic)
    origin : float
        data origin (according to selected voxel size)

    Notes
    -----
    Outputs a vtk, vtp or vti file that can opened in ParaView

    Examples
    --------
    `Click here
    <https://porespy.org/examples/io/reference/dict_to_vtk.html>`_
    to view online example.

    """
    try:
        from pyevtk.hl import imageToVTK
    except ModuleNotFoundError:
        msg = 'The pyevtk package can be installed with pip install pyevtk'
        raise ModuleNotFoundError(msg)
    vs = voxel_size
    for entry in data:
        if data[entry].dtype == bool:
            data[entry] = data[entry].astype(np.int8)
        if data[entry].flags["C_CONTIGUOUS"]:
            data[entry] = np.ascontiguousarray(data[entry])
    imageToVTK(filename, cellData=data, spacing=(vs, vs, vs), origin=origin)


def to_vtk(im, filename, divide=False, downsample=False, voxel_size=1, vox=False):
    r"""
    Converts an array to a vtk file.

    Parameters
    ----------
    im : 3D image
        The image of the porous material
    path : string
        Path to output file
    divide : bool
        vtk files can get very large, this option allows you for two output
        files, divided at z = half. This allows for large data sets to be
        imaged without loss of information
    downsample : bool
        very large images acan be downsampled to half the size in each
        dimension, this doubles the effective voxel size
    voxel_size : int
        The side length of the voxels (voxels  are cubic)
    vox : bool
        For an image that is binary (1's and 0's) this reduces the file size by
        using int8 format (can also be used to reduce file size when accuracy
        is not necessary ie: just visulization)

    Notes
    -----
    Outputs a vtk, vtp or vti file that can opened in paraview

    Examples
    --------
    `Click here
    <https://porespy.org/examples/io/reference/to_vtk.html>`_
    to view online example.

    """
    try:
        from pyevtk.hl import imageToVTK
    except ModuleNotFoundError:
        msg = 'The pyevtk package can be installed with pip install pyevtk'
        raise ModuleNotFoundError(msg)
    if len(im.shape) == 2:
        im = im[:, :, np.newaxis]
    if im.dtype == bool:
        vox = True
    if vox:
        im = im.astype(np.int8)
    vs = voxel_size
    if divide:
        split = np.round(im.shape[2] / 2).astype(np.int)
        im1 = im[:, :, 0:split]
        im2 = im[:, :, split:]
        imageToVTK(f"{filename}_1",
                   cellData={"im": np.ascontiguousarray(im1)},
                   spacing=(vs, vs, vs),)
        imageToVTK(f"{filename}_2",
                   origin=(0.0, 0.0, split * vs),
                   cellData={"im": np.ascontiguousarray(im2)},
                   spacing=(vs, vs, vs),)
    elif downsample:
        im = nd.interpolation.zoom(im, zoom=0.5, order=0, mode="reflect")
        imageToVTK(filename,
                   cellData={"im": np.ascontiguousarray(im)},
                   spacing=(2 * vs, 2 * vs, 2 * vs),)
    else:
        imageToVTK(filename,
                   cellData={"im": np.ascontiguousarray(im)},
                   spacing=(vs, vs, vs))


def to_palabos(im, filename, solid=0):
    r"""
    Converts an ndarray image to a text file that Palabos can read in as a
    geometry for Lattice Boltzmann simulations. Uses a Euclidean distance
    transform to identify solid voxels neighboring fluid voxels and labels
    them as the interface.

    Parameters
    ----------
    im : ndarray
        The image of the porous material
    filename : string
        Path to output file
    solid : int
        The value of the solid voxels in the image used to convert image to
        binary with all other voxels assumed to be fluid.

    Notes
    -----
    File produced contains 3 values: 2 = Solid, 1 = Interface, 0 = Pore
    Palabos will run the simulation applying the specified pressure drop from
    x = 0 to x = -1.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/io/reference/to_palabos.html>`_
    to view online example.

    """
    # Create binary image for fluid and solid phases
    bin_im = im == solid
    # Transform to integer for distance transform
    bin_im = bin_im.astype(int)
    # Distance Transform computes Euclidean distance in lattice units to
    # Nearest fluid for every solid voxel
    dt = nd.distance_transform_edt(bin_im)
    dt[dt > np.sqrt(2)] = 2
    dt[(dt > 0) * (dt <= np.sqrt(2))] = 1
    dt = dt.astype(int)
    # Write out data
    with open(filename, "w") as f:
        out_data = dt.flatten().tolist()
        f.write("\n".join(map(repr, out_data)))


def openpnm_to_im(
    network,
    pore_shape="sphere",
    throat_shape="cylinder",
    max_dim=None,
    rtol=0.1,
):
    r"""
    Generates voxel image from an OpenPNM network object.

    Parameters
    ----------
    network : OpenPNM GenericNetwork
        Network from which voxel image is to be generated
    pore_shape : str
        Shape of pores in the network, valid choices are "sphere", "cube"
    throat_shape : str
        Shape of throats in the network, valid choices are "cylinder", "cuboid"
    max_dim : int
        Number of voxels in the largest dimension of the network
    rtol : float
        Stopping criteria for finding the smallest voxel image such that
        further increasing the number of voxels in each dimension by 25% would
        improve the predicted porosity of the image by less that ``rtol``

    Returns
    -------
    im : ndarray
        Voxelated image corresponding to the given pore network model

    Notes
    -----
    (1) The generated voxelated image is labeled with 0s, 1s and 2s signifying
    solid phase, pores, and throats respectively.

    (2) If max_dim is not provided, the method calculates it such that the
    further increasing it doesn't change porosity by much.

    """
    return generate_voxel_image(
        network,
        pore_shape=pore_shape,
        throat_shape=throat_shape,
        max_dim=max_dim,
        rtol=rtol,
    )


def to_stl(im, filename, divide=False, downsample=False, voxel_size=1, vox=False):
    r"""
    Converts an array to an STL file.

    Parameters
    ----------
    im : 3D image
        The image of the porous material
    path : string
        Path to output file
    divide : bool
        vtk files can get very large, this option allows you for two output
        files, divided at z = half. This allows for large data sets to be
        imaged without loss of information
    downsample : bool
        very large images acan be downsampled to half the size in each
        dimension, this doubles the effective voxel size
    voxel_size : int
        The side length of the voxels (voxels  are cubic)
    vox : bool
        For an image that is binary (1's and 0's) this reduces the file size by
        using int8 format (can also be used to reduce file size when accuracy
        is not necessary ie: just visulization)

    Notes
    -----
    Outputs an STL file that can opened in Paraview

    Examples
    --------
    `Click here
    <https://porespy.org/examples/io/reference/to_stl.html>`_
    to view online example.

    """
    filename = sanitize_filename(filename, ext="stl", exclude_ext=True)
    if len(im.shape) == 2:
        im = im[:, :, np.newaxis]
    if im.dtype == bool:
        vox = True
    if vox:
        im = im.astype(np.int8)
    vs = voxel_size
    if divide:
        split = np.round(im.shape[2] / 2).astype(np.int)
        im1 = im[:, :, 0:split]
        im2 = im[:, :, split:]
        _save_stl(im1, vs, f"{filename}_1")
        _save_stl(im2, vs, f"{filename}_2")
    elif downsample:
        im = nd.interpolation.zoom(im, zoom=0.5, order=0, mode="reflect")
        _save_stl(im, vs * 2, filename)
    else:
        _save_stl(im, vs, filename)


def _save_stl(im, vs, filename):
    r"""
    Helper method to convert an array to an STL file.

    Parameters
    ----------
    im : 3D image
        The image of the porous material
    voxel_size : int
        The side length of the voxels (voxels are cubic)
    filename : string
        Path to output file

    """
    try:
        from stl import mesh
    except ModuleNotFoundError:
        msg = 'numpy-stl can be installed with pip install numpy-stl'
        ModuleNotFoundError(msg)
    im = np.pad(im, pad_width=10, mode="constant", constant_values=True)
    vertices, faces, norms, values = ms.marching_cubes(im)
    vertices *= vs
    # Export the STL file
    export = mesh.Mesh(np.zeros(faces.shape[0], dtype=mesh.Mesh.dtype))
    for i, f in enumerate(faces):
        for j in range(3):
            export.vectors[i][j] = vertices[f[j], :]
    export.save(f"{filename}.stl")


def to_paraview(im, filename, phase=2):
    r"""
    Converts an array to a paraview state file.

    Parameters
    ----------
    im : ndarray
        The image of the porous material.
    filename : str
        Path to output file.
    phase : str
        The desired phase of output image where phase = 0 represent the
        pore phase, phase = 1 represents the solid phase, and phase= 2 is
        the whole domain. The default value is 2.

    Notes
    -----
    Outputs an pvsm file that can opened in Paraview.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/io/reference/to_paraview.html>`_
    to view online example.

    """
    try:
        import imageio
    except ModuleNotFoundError:
        msg = "The imageio package can be installed with pip install imageio"
        raise ModuleNotFoundError(msg)
    try:
        import paraview.simple
    except ModuleNotFoundError:
        msg = ("The paraview python bindings must be installed using conda"
               " install -c conda-forge paraview, however this may require"
               " using a virtualenv since conflicts with other packages are"
               " common. This is why it is not explicitly included as a"
               " dependency in porespy.")
        raise ModuleNotFoundError(msg)
    data = im.astype("uint8")
    file = os.path.splitext(filename)[0]
    path = file + ".tiff"
    if len(im.shape) == 2:
        imageio.imwrite(path, np.array(data))
        view = "Slice"
        zshape = 0
        xshape = im.shape[1]
        yshape = im.shape[0]
    elif len(im.shape) == 3:
        imageio.volsave(path, np.array(data))
        view = "Volume"
        zshape = im.shape[0]
        xshape = im.shape[2]
        yshape = im.shape[1]
    maxshape = max(xshape, yshape)
    paraview.simple._DisableFirstRenderCameraReset()
    # Create a new 'TIFF Series Reader'
    dtiff = paraview.simple.TIFFSeriesReader(FileNames=[path])
    # Get active view
    renderView1 = paraview.simple.GetActiveViewOrCreate("RenderView")
    # Uncomment following to set a specific view size
    # renderView1.ViewSize = [1612, 552]
    # Get layout
    _ = paraview.simple.GetLayout()

    # Show data in view
    dtiffDisplay = paraview.simple.Show(dtiff, renderView1, "UniformGridRepresentation")

    # Get color transfer function/color map for 'TiffScalars'
    tiffScalarsLUT = paraview.simple.GetColorTransferFunction("TiffScalars")

    # Get opacity transfer function/opacity map for 'TiffScalars'
    tiffScalarsPWF = paraview.simple.GetOpacityTransferFunction("TiffScalars")

    # Trace defaults for the display properties.
    dtiffDisplay.Representation = view
    dtiffDisplay.ColorArrayName = ["POINTS", "Tiff Scalars"]
    dtiffDisplay.LookupTable = tiffScalarsLUT
    dtiffDisplay.OSPRayScaleArray = "Tiff Scalars"
    dtiffDisplay.OSPRayScaleFunction = "PiecewiseFunction"
    dtiffDisplay.SelectOrientationVectors = "None"
    dtiffDisplay.ScaleFactor = maxshape / 10 - 0.1
    dtiffDisplay.SelectScaleArray = "Tiff Scalars"
    dtiffDisplay.GlyphType = "Arrow"
    dtiffDisplay.GlyphTableIndexArray = "Tiff Scalars"
    dtiffDisplay.GaussianRadius = maxshape / 200 - 0.005
    dtiffDisplay.SetScaleArray = ["POINTS", "Tiff Scalars"]
    dtiffDisplay.ScaleTransferFunction = "PiecewiseFunction"
    dtiffDisplay.OpacityArray = ["POINTS", "Tiff Scalars"]
    dtiffDisplay.OpacityTransferFunction = "PiecewiseFunction"
    dtiffDisplay.DataAxesGrid = "GridAxesRepresentation"
    dtiffDisplay.PolarAxes = "PolarAxesRepresentation"
    dtiffDisplay.ScalarOpacityUnitDistance = 8.256564094912507
    dtiffDisplay.ScalarOpacityFunction = tiffScalarsPWF
    dtiffDisplay.IsosurfaceValues = [0.5]
    dtiffDisplay.SliceFunction = "Plane"

    shape = np.array([xshape, yshape, zshape])

    # Init the 'Plane' selected for 'SliceFunction'
    dtiffDisplay.SliceFunction.Origin = [xi / 2 - 0.5 for xi in shape]

    # Reset view to fit data
    renderView1.ResetCamera()

    # Changing interaction mode based on data extents
    # renderView1.InteractionMode = mode
    renderView1.CameraPosition = [
        xshape / 2 - 0.5,
        yshape / 2 - 0.5,
        4.6 * np.sqrt(np.sum(shape / 2 - 0.5)**2)
    ]
    renderView1.CameraFocalPoint = [xi / 2 - 0.5 for xi in shape]

    # Get the material library
    _ = paraview.simple.GetMaterialLibrary()

    # Show color bar/color legend
    dtiffDisplay.SetScalarBarVisibility(renderView1, True)

    # Update the view to ensure updated data information
    renderView1.Update()

    # Saving camera placements for all active views
    # Current camera placement for renderView1
    # renderView1.InteractionMode = mode
    renderView1.CameraPosition = [
        xshape / 2 - 0.5,
        yshape / 2 - 0.5,
        4.6 * np.sqrt(np.sum(shape / 2 - 0.5)**2)
    ]
    renderView1.CameraFocalPoint = [xi / 2 - 0.5 for xi in shape]
    renderView1.CameraParallelScale = np.sqrt(np.sum(shape / 2 - 0.5)**2)

    # Uncomment the following to render all views
    # RenderAllViews()
    # Alternatively, if you want to write images, you can use SaveScreenshot(...).
    threshold1 = paraview.simple.Threshold(Input=dtiff)
    threshold1.Scalars = ["POINTS", "Tiff Scalars"]
    if phase == 0:
        threshold_range = [0.5, 1]
    elif phase == 1:
        threshold_range = [0, 0.5]
    else:
        threshold_range = [0, 1]
    threshold1.ThresholdRange = threshold_range

    # Show data in view
    _ = paraview.simple.Show(
        threshold1, renderView1, "UnstructuredGridRepresentation"
    )

    # Hide data in view
    paraview.simple.Hide(dtiff, renderView1)

    paraview.simple.SaveState(file + ".pvsm")


def open_paraview(filename=None, im=None, **kwargs):
    r"""
    Open a paraview state file or image directly in paraview.

    Parameters
    ----------
    filename : str
        Path to input state file.
    im : ndarray
        An image to open directly.  If no filename given, then this image is
        sent to ``to_paraview`` and a state file is created with a random name.
        Any additional keyword arguments are sent to ``to_paraview``.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/io/reference/open_paraview.html>`_
    to view online example.

    """
    if filename is None:
        from datetime import datetime
        now = datetime.now()
        filename = now.strftime("%d-%m-%Y_%H-%M-%S")
        to_paraview(im=im, filename=filename, **kwargs)
    file = os.path.splitext(filename)[0]
    statefile = file + ".pvsm"
    # paraview_path = "paraview.exe"
    paraview_path = "paraview"
    subprocess.Popen([paraview_path, statefile])


def spheres_to_comsol(filename, im=None, centers=None, radii=None):
    r"""
    Exports a sphere pack into a Comsol geometry file.

    An image containing spheres can be specified.  Alternatively as list of
    ``centers`` and ``radii`` can be given if known.

    Parameters
    ----------
    filename : string or path object
        Location and namge to output file
    im : ndarray (optional)
        A voxel image containing spheres indicated by non-zeros values.
        Spheres can be generated using a variety of methods and can overlap.
        The sphere centers and radii are found as the peaks in the
        distance transform.  If ``im`` is not supplied, then ``centers`` and
        ``radii`` must be given.
    centers : array_like (optional)
        An array (Ns, 3) of the spheres centers where Ns is the number of
        spheres.  This must be specified if ``im`` is not suppplied.
    radii : array_like (optional)
        An Ns length array of the spheres's. This must be specified if ``im``
        is not suppplied.

    Notes
    -----
    If ``im`` is given then some image analysis is performed to find sphere
    centers so it may not perfectly represent the spheres in the original
    image. This is especially true for overlapping sphere and spheres extending
    beyond the edge of the image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/io/reference/spheres_to_comsol.html>`_
    to view online example.

    """
    from ._comsol import _save_to_comsol
    if im is not None:
        if im.ndim != 3:
            raise Exception('Image must be 3D.')
        dt = edt(im > 0)
        dt2 = nd.gaussian_filter(dt, sigma=0.1)
        peaks = (im > 0)*(nd.maximum_filter(dt2, footprint=ball(3)) == dt)
        peaks = reduce_peaks(peaks)
        centers = np.vstack(np.where(peaks)).T
        radii = dt[tuple(centers.T)].astype(int)
    _save_to_comsol(filename, centers, radii)



import logging
import numpy as np
import scipy.ndimage as spim
from skimage.morphology import disk, ball
from edt import edt
from porespy.tools import extend_slice
from porespy import settings
from porespy.tools import get_tqdm, make_contiguous
from porespy.metrics import region_surface_areas, region_interface_areas
from porespy.metrics import region_volumes


__all__ = [
    "regions_to_network",
]


tqdm = get_tqdm()
logger = logging.getLogger(__name__)


def regions_to_network(regions, phases=None, voxel_size=1, accuracy='standard'):
    r"""
    Analyzes an image that has been partitioned into pore regions and extracts
    the pore and throat geometry as well as network connectivity.

    Parameters
    ----------
    regions : ndarray
        An image of the material partitioned into individual regions.
        Zeros in this image are ignored.
    phases : ndarray, optional
        An image indicating to which phase each voxel belongs. The returned
        network contains a 'pore.phase' array with the corresponding value.
        If not given a value of 1 is assigned to every pore.
    voxel_size : scalar (default = 1)
        The resolution of the image, expressed as the length of one side of a
        voxel, so the volume of a voxel would be **voxel_size**-cubed.
    accuracy : string
        Controls how accurately certain properties are calculated. Options are:

        'standard' (default)
            Computes the surface areas and perimeters by simply counting
            voxels.  This is *much* faster but does not properly account
            for the rough, voxelated nature of the surfaces.
        'high'
            Computes surface areas using the marching cube method, and
            perimeters using the fast marching method.  These are substantially
            slower but better account for the voxelated nature of the images.

    Returns
    -------
    net : dict
        A dictionary containing all the pore and throat size data, as well as
        the network topological information.  The dictionary names use the
        OpenPNM convention (i.e. 'pore.coords', 'throat.conns').

    Notes
    -----
    The meaning of each of the values returned in ``net`` are outlined below:

    'pore.region_label'
        The region labels corresponding to the watershed extraction. The
        pore indices and regions labels will be offset by 1, so pore 0
        will be region 1.
    'throat.conns'
        An *Nt-by-2* array indicating which pores are connected to each other
    'pore.region_label'
        Mapping of regions in the watershed segmentation to pores in the
        network
    'pore.local_peak'
        The coordinates of the location of the maxima of the distance transform
        performed on the pore region in isolation
    'pore.global_peak'
        The coordinates of the location of the maxima of the distance transform
        performed on the full image
    'pore.geometric_centroid'
        The center of mass of the pore region as calculated by
        ``skimage.measure.center_of_mass``
    'throat.global_peak'
        The coordinates of the location of the maxima of the distance transform
        performed on the full image
    'pore.region_volume'
        The volume of the pore region computed by summing the voxels
    'pore.volume'
        The volume of the pore found by as volume of a mesh obtained from the
        marching cubes algorithm
    'pore.surface_area'
        The surface area of the pore region as calculated by either counting
        voxels or using the fast marching method to generate a tetramesh (if
        ``accuracy`` is set to ``'high'``.)
    'throat.cross_sectional_area'
        The cross-sectional area of the throat found by either counting
        voxels or using the fast marching method to generate a tetramesh (if
        ``accuracy`` is set to ``'high'``.)
    'throat.perimeter'
        The perimeter of the throat found by counting voxels on the edge of
        the region defined by the intersection of two regions.
    'pore.inscribed_diameter'
        The diameter of the largest sphere inscribed in the pore region. This
        is found as the maximum of the distance transform on the region in
        isolation.
    'pore.extended_diameter'
        The diamter of the largest sphere inscribed in overal image, which
        can extend outside the pore region. This is found as the local maximum
        of the distance transform on the full image.
    'throat.inscribed_diameter'
        The diameter of the largest sphere inscribed in the throat.  This
        is found as the local maximum of the distance transform in the area
        where to regions meet.
    'throat.total_length'
        The length between pore centered via the throat center
    'throat.direct_length'
        The length between two pore centers on a straight line between them
        that does not pass through the throat centroid.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/networks/reference/regions_to_network.html>`_
    to view online example.

    """
    logger.info('Extracting pore/throat information')

    im = make_contiguous(regions)
    struc_elem = disk if im.ndim == 2 else ball
    voxel_size = float(voxel_size)
    if phases is None:
        phases = (im > 0).astype(int)
    if im.size != phases.size:
        raise Exception('regions and phase are different sizes, probably ' +
                        'because boundary regions were not added to phases')
    dt = np.zeros_like(phases, dtype="float32")  # since edt returns float32
    for i in np.unique(phases[phases.nonzero()]):
        dt += edt(phases == i)

    # Get 'slices' into im for each pore region
    slices = spim.find_objects(im)

    # Initialize arrays
    Ps = np.arange(1, np.amax(im)+1)
    Np = np.size(Ps)
    p_coords_cm = np.zeros((Np, im.ndim), dtype=float)
    p_coords_dt = np.zeros((Np, im.ndim), dtype=float)
    p_coords_dt_global = np.zeros((Np, im.ndim), dtype=float)
    p_volume = np.zeros((Np, ), dtype=float)
    p_dia_local = np.zeros((Np, ), dtype=float)
    p_dia_global = np.zeros((Np, ), dtype=float)
    p_label = np.zeros((Np, ), dtype=int)
    p_area_surf = np.zeros((Np, ), dtype=int)
    p_phase = np.zeros((Np, ), dtype=int)
    # The number of throats is not known at the start, so lists are used
    # which can be dynamically resized more easily.
    t_conns = []
    t_dia_inscribed = []
    t_area = []
    t_perimeter = []
    t_coords = []

    # Start extracting size information for pores and throats
    msg = "Extracting pore and throat properties"
    for i in tqdm(Ps, desc=msg, **settings.tqdm):
        pore = i - 1
        if slices[pore] is None:
            continue
        s = extend_slice(slices[pore], im.shape)
        sub_im = im[s]
        sub_dt = dt[s]
        pore_im = sub_im == i
        padded_mask = np.pad(pore_im, pad_width=1, mode='constant')
        pore_dt = edt(padded_mask)
        s_offset = np.array([i.start for i in s])
        p_label[pore] = i
        p_coords_cm[pore, :] = spim.center_of_mass(pore_im) + s_offset
        temp = np.vstack(np.where(pore_dt == pore_dt.max()))[:, 0]
        p_coords_dt[pore, :] = temp + s_offset
        p_phase[pore] = (phases[s]*pore_im).max()
        temp = np.vstack(np.where(sub_dt == sub_dt.max()))[:, 0]
        p_coords_dt_global[pore, :] = temp + s_offset
        p_volume[pore] = np.sum(pore_im, dtype=np.int64)
        p_dia_local[pore] = 2*np.amax(pore_dt)
        p_dia_global[pore] = 2*np.amax(sub_dt)
        # The following is overwritten if accuracy is set to 'high'
        p_area_surf[pore] = np.sum(pore_dt == 1, dtype=np.int64)
        im_w_throats = spim.binary_dilation(input=pore_im, structure=struc_elem(1))
        im_w_throats = im_w_throats*sub_im
        Pn = np.unique(im_w_throats)[1:] - 1
        for j in Pn:
            if j > pore:
                t_conns.append([pore, j])
                vx = np.where(im_w_throats == (j + 1))
                t_dia_inscribed.append(2*np.amax(sub_dt[vx]))
                # The following is overwritten if accuracy is set to 'high'
                t_perimeter.append(np.sum(sub_dt[vx] < 2, dtype=np.int64))
                # The following is overwritten if accuracy is set to 'high'
                t_area.append(np.size(vx[0]))
                p_area_surf[pore] -= np.size(vx[0])
                t_inds = tuple([i+j for i, j in zip(vx, s_offset)])
                temp = np.where(dt[t_inds] == np.amax(dt[t_inds]))[0][0]
                t_coords.append(tuple([t_inds[k][temp] for k in range(im.ndim)]))

    # Clean up values
    p_coords = p_coords_cm
    Nt = len(t_dia_inscribed)  # Get number of throats
    if im.ndim == 2:  # If 2D, add 0's in 3rd dimension
        p_coords = np.vstack((p_coords_cm.T, np.zeros((Np, )))).T
        t_coords = np.vstack((np.array(t_coords).T, np.zeros((Nt, )))).T

    net = {}
    ND = im.ndim
    # Define all the fundamental stuff
    net['throat.conns'] = np.array(t_conns)
    net['pore.coords'] = np.array(p_coords)*voxel_size
    net['pore.all'] = np.ones_like(net['pore.coords'][:, 0], dtype=bool)
    net['throat.all'] = np.ones_like(net['throat.conns'][:, 0], dtype=bool)
    net['pore.region_label'] = np.array(p_label)
    net['pore.phase'] = np.array(p_phase, dtype=int)
    net['throat.phases'] = net['pore.phase'][net['throat.conns']]
    V = np.copy(p_volume)*(voxel_size**ND)
    net['pore.region_volume'] = V  # This will be an area if image is 2D
    f = 3/4 if ND == 3 else 1.0
    net['pore.equivalent_diameter'] = 2*(V/np.pi * f)**(1/ND)
    # Extract the geometric stuff
    net['pore.local_peak'] = np.copy(p_coords_dt)*voxel_size
    net['pore.global_peak'] = np.copy(p_coords_dt_global)*voxel_size
    net['pore.geometric_centroid'] = np.copy(p_coords_cm)*voxel_size
    net['throat.global_peak'] = np.array(t_coords)*voxel_size
    net['pore.inscribed_diameter'] = np.copy(p_dia_local)*voxel_size
    net['pore.extended_diameter'] = np.copy(p_dia_global)*voxel_size
    net['throat.inscribed_diameter'] = np.array(t_dia_inscribed)*voxel_size
    P12 = net['throat.conns']
    PT1 = np.sqrt(np.sum(((p_coords[P12[:, 0]]-t_coords)*voxel_size)**2,
                         axis=1))
    PT2 = np.sqrt(np.sum(((p_coords[P12[:, 1]]-t_coords)*voxel_size)**2,
                         axis=1))
    net['throat.total_length'] = PT1 + PT2
    PT1 = PT1-p_dia_local[P12[:, 0]]/2*voxel_size
    PT2 = PT2-p_dia_local[P12[:, 1]]/2*voxel_size
    dist = (p_coords[P12[:, 0]] - p_coords[P12[:, 1]])*voxel_size
    net['throat.direct_length'] = np.sqrt(np.sum(dist**2, axis=1, dtype=np.int64))
    net['throat.perimeter'] = np.array(t_perimeter)*voxel_size
    if (accuracy == 'high') and (im.ndim == 2):
        msg = "accuracy='high' only available in 3D, reverting to 'standard'"
        logger.warning(msg)
        accuracy = 'standard'
    if (accuracy == 'high'):
        net['pore.volume'] = region_volumes(regions=im, mode='marching_cubes')
        areas = region_surface_areas(regions=im, voxel_size=voxel_size)
        net['pore.surface_area'] = areas
        interface_area = region_interface_areas(regions=im, areas=areas,
                                                voxel_size=voxel_size)
        A = interface_area.area
        net['throat.cross_sectional_area'] = A
        net['throat.equivalent_diameter'] = (4*A/np.pi)**(1/2)
    else:
        net['pore.volume'] = np.copy(p_volume)*(voxel_size**ND)
        net['pore.surface_area'] = np.copy(p_area_surf)*(voxel_size**2)
        A = np.array(t_area)*(voxel_size**2)
        net['throat.cross_sectional_area'] = A
        net['throat.equivalent_diameter'] = (4*A/np.pi)**(1/2)

    return net



import logging
import numpy as np
import scipy.ndimage as spim
from porespy.tools import extend_slice, ps_round
from porespy.tools import _check_for_singleton_axes, Results
from porespy.tools import mesh_region
from skimage import measure
from porespy.tools import get_tqdm
from porespy import settings


__all__ = [
    "mesh_surface_area",
    "mesh_volume",
    "region_interface_areas",
    "region_surface_areas",
    "region_volumes",
]


tqdm = get_tqdm()
logger = logging.getLogger(__name__)


def region_volumes(regions, mode='marching_cubes'):
    r"""
    Compute volume of each labelled region in an image

    Parameters
    ----------
    regions : ndarray
        An image with labelled regions
    mode : string
        Controls the method used. Options are:

        'marching_cubes' (default)
            Finds a mesh for each region using the marching cubes algorithm
            from ``scikit-image``, then finds the volume of the mesh using the
            ``trimesh`` package.

        'voxel'
            Calculates the region volume as the sum of voxels within each
            region.

    Returns
    -------
    volumes : ndarray
        An array of shape [N by 1] where N is the number of labelled regions
        in the image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/mesh_volumes.html>`_
    to view online example.

    """
    slices = spim.find_objects(regions)
    vols = np.zeros([len(slices), ])
    msg = "Computing region volumes".ljust(60)
    for i, s in enumerate(tqdm(slices, desc=msg, **settings.tqdm)):
        region = regions[s] == (i + 1)
        if mode == 'marching_cubes':
            vols[i] = mesh_volume(region)
        elif mode.startswith('voxel'):
            vols[i] = region.sum(dtype=np.int64)
    return vols


def mesh_volume(region):
    r"""
    Compute the volume of a single region by meshing it

    Parameters
    ----------
    region : ndarray
        An image with a single region labelled as ``True`` (or > 0)

    Returns
    -------
    volume : float
        The volume of the region computed by applyuing the marching cubes
        algorithm to the region, then finding the mesh volume using the
        ``trimesh`` package.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/mesh_volume.html>`_
    to view online example.

    """
    try:
        from trimesh import Trimesh
    except ModuleNotFoundError:
        msg = 'The trimesh package can be installed with pip install trimesh'
        raise ModuleNotFoundError(msg)
    mc = mesh_region(region > 0)
    m = Trimesh(vertices=mc.verts, faces=mc.faces, vertex_normals=mc.norm)
    if m.is_watertight:
        vol = np.abs(m.volume)
    else:
        vol = np.nan
    return vol


def region_surface_areas(regions, voxel_size=1, strel=None):
    r"""
    Extract the surface area of each region in a labeled image.

    Optionally, it can also find the the interfacial area between all
    adjoining regions.

    Parameters
    ----------
    regions : ndarray
        An image of the pore space partitioned into individual pore regions.
        Note that zeros in the image will not be considered for area
        calculation.
    voxel_size : scalar
        The resolution of the image, expressed as the length of one side of a
        voxel, so the volume of a voxel would be **voxel_size**-cubed.  The
        default is 1.
    strel : array_like
        The structuring element used to blur the region.  If not provided,
        then a spherical element (or disk) with radius 1 is used.  See the
        docstring for ``mesh_region`` for more details, as this argument is
        passed to there.

    Returns
    -------
    areas : list
        A list containing the surface area of each region, offset by 1, such
        that the surface area of region 1 is stored in element 0 of the list.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/region_surface_areas.html>`_
    to view online example.

    """
    logger.info('Finding surface area of each region')
    im = regions
    if strel is None:
        strel = ps_round(1, im.ndim, smooth=False)
    # Get 'slices' into im for each pore region
    slices = spim.find_objects(im)
    # Initialize arrays
    Ps = np.arange(1, np.amax(im) + 1)
    sa = np.zeros_like(Ps, dtype=float)
    # Start extracting marching cube area from im
    msg = "Computing region surface area".ljust(60)
    for i in tqdm(Ps, desc=msg, **settings.tqdm):
        reg = i - 1
        if slices[reg] is not None:
            s = extend_slice(slices[reg], im.shape)
            sub_im = im[s]
            mask_im = sub_im == i
            mesh = mesh_region(region=mask_im, strel=strel)
            sa[reg] = mesh_surface_area(mesh)
    result = sa * voxel_size**2
    return result


def mesh_surface_area(mesh=None, verts=None, faces=None):
    r"""
    Calculate the surface area of a meshed region

    Parameters
    ----------
    mesh : tuple
        The tuple returned from the ``mesh_region`` function
    verts : array
        An N-by-ND array containing the coordinates of each mesh vertex
    faces : array
        An N-by-ND array indicating which elements in ``verts`` form a mesh
        element.

    Returns
    -------
    surface_area : float
        The surface area of the mesh, calculated by
        ``skimage.measure.mesh_surface_area``

    Notes
    -----
    This function simply calls ``scikit-image.measure.mesh_surface_area``, but
    it allows for the passing of the ``mesh`` tuple returned by the
    ``mesh_region`` function, entirely for convenience.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/mesh_surface_area.html>`_
    to view online example.

    """
    if mesh:
        verts = mesh.verts
        faces = mesh.faces
    else:
        if (verts is None) or (faces is None):
            raise Exception('Either mesh or verts and faces must be given')
    surface_area = measure.mesh_surface_area(verts, faces)
    return surface_area


def region_interface_areas(regions, areas, voxel_size=1, strel=None):
    r"""
    Calculate the interfacial area between all pairs of adjecent regions

    Parameters
    ----------
    regions : ndarray
        An image of the pore space partitioned into individual pore regions.
        Note that zeros in the image will not be considered for area
        calculation.
    areas : array_like
        A list containing the areas of each regions, as determined by
        ``region_surface_area``.  Note that the region number and list index
        are offset by 1, such that the area for region 1 is stored in
        ``areas[0]``.
    voxel_size : scalar
        The resolution of the image, expressed as the length of one side of a
        voxel, so the volume of a voxel would be **voxel_size**-cubed.  The
        default is 1.
    strel : array_like
        The structuring element used to blur the region.  If not provided,
        then a spherical element (or disk) with radius 1 is used.  See the
        docstring for ``mesh_region`` for more details, as this argument is
        passed to there.

    Returns
    -------
    areas : Results object
        A custom object with the following data added as named attributes:

        'conns'
            An N-regions by 2 array with each row containing the region
            number of an adjacent pair of regions.  For instance, if
            ``conns[0, 0]`` is 0 and ``conns[0, 1]`` is 5, then row 0 of
            ``area`` contains the interfacial area shared by regions 0 and 5.

        'area'
            The area calculated for each pair of regions in ``conns``

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/region_interface_areas.html>`_
    to view online example.

    """
    logger.info('Finding interfacial areas between each region')
    im = regions
    _check_for_singleton_axes(im)
    ball = ps_round(1, im.ndim, smooth=False)
    if strel is None:
        strel = np.copy(ball)
    # Get 'slices' into im for each region
    slices = spim.find_objects(im)
    # Initialize arrays
    Ps = np.arange(1, np.amax(im) + 1)
    sa = np.zeros_like(Ps, dtype=float)
    sa_combined = []  # Difficult to preallocate since number of conns unknown
    cn = []
    # Start extracting area from im
    msg = "Computing interfacial area between regions".ljust(60)
    for i in tqdm(Ps, desc=msg, **settings.tqdm):
        reg = i - 1
        if slices[reg] is not None:
            s = extend_slice(slices[reg], im.shape)
            sub_im = im[s]
            mask_im = sub_im == i
            sa[reg] = areas[reg]
            im_w_throats = spim.binary_dilation(input=mask_im,
                                                structure=ball)
            im_w_throats = im_w_throats * sub_im
            Pn = np.unique(im_w_throats)[1:] - 1
            for j in Pn:
                if j > reg:
                    cn.append([reg, j])
                    merged_region = im[(min(slices[reg][0].start,
                                            slices[j][0].start)):
                                       max(slices[reg][0].stop,
                                           slices[j][0].stop),
                                       (min(slices[reg][1].start,
                                            slices[j][1].start)):
                                       max(slices[reg][1].stop,
                                           slices[j][1].stop)]
                    merged_region = ((merged_region == reg + 1)
                                     + (merged_region == j + 1))
                    mesh = mesh_region(region=merged_region, strel=strel)
                    sa_combined.append(mesh_surface_area(mesh))
    # Interfacial area calculation
    cn = np.array(cn)
    ia = 0.5 * (sa[cn[:, 0]] + sa[cn[:, 1]] - sa_combined)
    ia[ia <= 0] = 1
    result = Results()
    result.conns = cn
    result.area = ia * voxel_size**2
    return result
